"""
Define accessor routines.

This is inspired by xarray accessors.
"""

from __future__ import annotations

import warnings

# from itertools import chain
# from .utils import get_tqdm_calc as get_tqdm
# from .utils import parallel_map_attr, parallel_map_call
from typing import TYPE_CHECKING, Callable, Generic, overload

from module_utilities import cached
from module_utilities.typing import C_prop, R, S

if TYPE_CHECKING:
    from typing import Literal

    from module_utilities.cached import CachedProperty


class AccessorRegistrationWarning(Warning):
    """Warning for conflicts in accessor registration."""


class _CachedAccessorSingle(Generic[S, R]):
    """
    Custom property-like object (descriptor).
    this gets created once on call
    """

    def __init__(self, name: str, accessor: C_prop[S, R]) -> None:
        self.__doc__ = accessor.__doc__

        self._name = name
        self._accessor = accessor

    @overload
    def __get__(self, obj: None, cls: type[S] | None = None) -> C_prop[S, R]: ...

    @overload
    def __get__(self, obj: S, cls: type[S] | None = None) -> R: ...

    def __get__(self, obj: S | None, cls: type[S] | None = None) -> C_prop[S, R] | R:
        if obj is None:
            return self._accessor
        try:
            accessor_obj = self._accessor(obj)
        except AttributeError as err:
            # __getattr__ on data object will swallow any AttributeErrors
            # raised when initializing the accessor, so we need to raise as
            # something else (GH933):
            raise RuntimeError("error initializing %r accessor." % self._name) from err
        # Replace the property with the accessor object. Inspired by:
        # http://www.pydanny.com/cached-property.html
        # We need to use object.__setattr__ because we overwrite __setattr__ on
        # AttrAccessMixin.
        object.__setattr__(obj, self._name, accessor_obj)
        return accessor_obj


def _cachedaccessorcleared(name: str, accessor: C_prop[S, R]) -> CachedProperty[S, R]:
    """
    Wrap accessor in a cached property

    This creates an property in the parent class with name `name`
    from `accessor(self)`.
    This only gets initialized when called

    Notes
    -----
    This gets deleted if self._cache is cleared.
    If you only want to create it once, then use the Single access wrapper below
    """

    # @cached.prop(key=name)
    # @wraps(accessor)
    # def _get_prop(self):
    #     return accessor(self)

    # return _get_prop
    return cached.prop(key=name)(accessor)


@overload
def _cachedaccessorwrapper(
    name: str, accessor: C_prop[S, R], *, single_create: Literal[True]
) -> _CachedAccessorSingle[S, R]: ...


@overload
def _cachedaccessorwrapper(
    name: str, accessor: C_prop[S, R], *, single_create: Literal[False] = ...
) -> CachedProperty[S, R]: ...


@overload
def _cachedaccessorwrapper(
    name: str, accessor: C_prop[S, R], *, single_create: bool
) -> _CachedAccessorSingle[S, R] | CachedProperty[S, R]: ...


def _cachedaccessorwrapper(
    name: str, accessor: C_prop[S, R], *, single_create: bool = False
) -> _CachedAccessorSingle[S, R] | CachedProperty[S, R]:
    if single_create:
        return _CachedAccessorSingle(name, accessor)
    return _cachedaccessorcleared(name, accessor)


class AccessorMixin:  # (Generic[S, R]):
    """Mixin for accessor."""

    @classmethod
    def _register_accessor(
        cls, name: str, accessor: C_prop[S, R], single_create: bool = False
    ) -> None:
        """Most general accessor"""
        if hasattr(cls, name):
            warnings.warn(
                f"registration of accessor {accessor!r} under name {name!r} for type {cls!r} is "
                "overriding a preexisting attribute with the same name.",
                AccessorRegistrationWarning,
                stacklevel=2,
            )
        setattr(
            cls,
            name,
            _cachedaccessorwrapper(name, accessor, single_create=single_create),
        )

    # @classmethod
    # def register_accessor_flat(cls, names: str, single_create: bool=True, subattr="flat"):
    #     """Helper class to register stuff to top of CollectionPhases"""

    #     if isinstance(names, str):
    #         names = [names]
    #     for name in names:
    #         cls._register_accessor(
    #             name=name,
    #             accessor=attrgetter(".".join((subattr, name))),
    #             single_create=single_create,
    #         )

    @classmethod
    def register_accessor(
        cls, name: str, accessor: C_prop[S, R], single_create: bool = False
    ) -> None:
        """
        Register a property `name` to `class` of type `accessor(self)`

        Examples
        --------
        >>> class parent(AccessorMixin):
        ...     pass
        >>> class hello(AccessorMixin):
        ...     def __init__(self, parent):
        ...         self._parent = parent
        ...
        ...     def there(self):
        ...         return f"{type(self._parent)}"

        >>> parent.register_accessor("hello", hello)
        >>> x = parent()
        >>> x.hello.there()
        "<class 'lnpy.extensions.parent'>"
        """
        cls._register_accessor(name, accessor, single_create=single_create)

    @classmethod
    def decorate_accessor(
        cls, name: str, single_create: bool = False
    ) -> Callable[[C_prop[S, R]], C_prop[S, R]]:
        """
        Register a property `name` to `class` of type `accessor(self)`.

        Examples
        --------
        >>> class parent(AccessorMixin):
        ...     pass

        >>> @parent.decorate_accessor("hello")
        ... class hello(AccessorMixin):
        ...     def __init__(self, parent):
        ...         self._parent = parent
        ...
        ...     def there(self):
        ...         return f"{type(self._parent)}"

        >>> x = parent()
        >>> x.hello.there()
        "<class 'lnpy.extensions.parent'>"
        """

        def decorator(accessor: C_prop[S, R]) -> C_prop[S, R]:
            cls.register_accessor(name, accessor, single_create=single_create)
            return accessor

        return decorator


################################################################################
# List access
# class _CallableListResultsCache:
#     """if items of collection accessor are callable, then"""

#     def __init__(self, parent, items, desc=None):
#         self.parent = parent
#         self.items = items
#         self.desc = desc
#         self._cache = {}
#         self._use_joblib = getattr(self.parent, "_USE_JOBLIB_", False)

#     @cached.meth
#     def __call__(self, *args, **kwargs):
#         # get value
#         seq = get_tqdm(self.items, desc=self.desc)
#         # results = [x(*args, **kwargs) for x in seq]
#         results = parallel_map_call(seq, self._use_joblib, *args, **kwargs)
#         if hasattr(self.parent, "wrap_list_results"):
#             results = self.parent.wrap_list_results(results)
#         return results


# class _CallableListResultsNoCache:
#     """if items of collection accessor are callable, then"""

#     def __init__(self, parent, items, desc=None):
#         self.parent = parent
#         self.items = items
#         self.desc = desc
#         self._cache = {}
#         self._use_joblib = getattr(self.parent, "_USE_JOBLIB_", False)

#     def __call__(self, *args, **kwargs):
#         # get value
#         seq = get_tqdm(self.items, desc=self.desc)
#         # results = [x(*args, **kwargs) for x in seq]
#         results = parallel_map_call(seq, self._use_joblib, *args, **kwargs)
#         if hasattr(self.parent, "wrap_list_results"):
#             results = self.parent.wrap_list_results(results)
#         return results


# def _CallableListResults(parent, items, use_cache=False, desc=None):
#     if use_cache:
#         cls = _CallableListResultsCache
#     else:
#         cls = _CallableListResultsNoCache

#     return cls(parent=parent, items=items, desc=desc)


# class _ListAccessor:
#     def __init__(self, parent, items, cache_list=None):
#         self.parent = parent
#         self.items = items

#         if cache_list is None:
#             cache_list = []
#         self._cache_list = cache_list
#         self._cache = {}
#         self._use_joblib = getattr(self.parent, "_USE_JOBLIB_", False)

#     def __getattr__(self, attr):
#         if attr in self._cache:
#             # print('using cache')
#             return self._cache[attr]

#         try:
#             use_cache = attr in self._cache_list
#             if callable(getattr(self.items[0], attr)):
#                 seq = [getattr(x, attr) for x in self.items]
#                 result = _CallableListResults(
#                     self.parent, seq, use_cache=use_cache, desc=attr
#                 )
#             else:
#                 seq = get_tqdm(self.items, desc=attr)
#                 # result = [getattr(x, attr) for x in seq]
#                 result = parallel_map_attr(attr, items=seq, use_joblib=self._use_joblib)
#                 if hasattr(self.parent, "wrap_list_results"):
#                     result = self.parent.wrap_list_results(result)

#         except Exception:
#             raise AttributeError(f"no attribute {attr} found")

#         # do caching?
#         if use_cache:
#             # print('caching')
#             self._cache[attr] = result
#         return result

#     def __getitem__(self, idx):
#         return self.items[idx]

#     def __dir__(self):
#         heritage = dir(super(self.__class__, self))
#         # hide = []
#         x = self.items[0]
#         show = [
#             k
#             for k in chain(
#                 self.__dict__.keys(),
#                 dir(x)
#                 # self.__class__.__dict__.keys()
#                 # x.__dict__.keys(),
#                 # x.__class__.__dict__.keys()
#             )
#         ]  # if k not in hide]
#         return sorted(heritage + show)


# def _CachedListPropertyWrapper(name, cache_list=None):
#     """Get top level property from items"""

#     if cache_list is not None and name in cache_list:
#         cache = True
#     else:
#         cache = False

#     # @cached.prop(key=name)
#     def _get_prop(self):
#         results = [getattr(x, name) for x in self]
#         if callable(results[0]):
#             results = _CallableListResults(self, results)
#         else:
#             if hasattr(self, "wrap_list_results"):
#                 results = self.wrap_list_results(results)
#         return results

#     if cache:
#         _get_prop = cached.prop(key=name)(_get_prop)
#     else:
#         _get_prop = property(_get_prop)

#     return _get_prop


# def _CachedListAccessorWrapper(name, cache_list=None):
#     """Wrap List accessor in cached property"""

#     @cached.prop(key=name)
#     def _get_prop(self):
#         return _ListAccessor(
#             self, [getattr(x, name) for x in self], cache_list=cache_list
#         )

#     return _get_prop


# class ListAccessorMixin:
#     """Mixin for list accessor."""

#     @classmethod
#     def _register_listaccessor(cls, names, accessor_wrapper, **kwargs):
#         """Most general accessor"""
#         if isinstance(names, str):
#             names = [names]
#         for name in names:
#             if hasattr(cls, name):
#                 warnings.warn(
#                     "registration of name {!r} for type {!r} is "
#                     "overriding a preexisting attribute with the same name.".format(
#                         name, cls
#                     ),
#                     AccessorRegistrationWarning,
#                     stacklevel=2,
#                 )
#             setattr(cls, name, accessor_wrapper(name, **kwargs))

#     @classmethod
#     def register_listaccessor(cls, names, cache_list=None):
#         """
#         Create an accessor to elements of parent class.

#         Examples
#         --------
#         class Prop(object):
#             def __init__(self, prop)
#                 self.prop = prop
#         class Child(object):
#             def __init__(self, a, b):
#                 self.a = Prop(a)
#         class Wrapper(ListAccessorMixin):
#             def __init__(self, items):
#                 self.items = items
#             def __getitem__(self, idx):
#                 # must have getitem for this to work
#                 return self.items[idx]
#             def wrap_list_results(self, result):
#                 #anython special to do with resultsgoes here
#                 return results

#         wrapper.register_listaccessor('a')
#         >>> x = [Child(i) for i in range(3)]
#         >>> w = Wrapper(x)
#         >>> w.a.prop
#         [1, 2, 3]
#         """
#         return cls._register_listaccessor(
#             names, accessor_wrapper=_CachedListAccessorWrapper, cache_list=cache_list
#         )

#     @classmethod
#     def register_listproperty(cls, names, cache_list=None):
#         """
#         Create an accessor to property

#         This is to give top level access to stuff

#         Examples
#         --------
#         class Child(object):
#             def __init__(self, a):
#                 self.a = a
#             def b(self, x):
#                 return self.a, x
#         class Wrapper(ListAccessorMixin):
#             def __init__(self, items):
#                 self.items = items

#             def __getitem__(self, idx):
#                 # must have getitem for this to work
#                 return self.items[idx]

#             def wrap_list_results(self, result):
#                 #anython special to do with resultsgoes here
#                 return results

#         Wrapper.register_listproperty(['a', 'b'])
#         >>> x = [Child(i) for i in range(3)]
#         >>> w = Wrapper(x)
#         >>> w.a
#         [1, 2, 3]
#         >>> w.b("a")
#         [(1,'a'), (2,'a'), (3,'a')]
#         """
#         return cls._register_listaccessor(
#             names, accessor_wrapper=_CachedListPropertyWrapper, cache_list=cache_list
#         )


# def _decorate_listaccessor(names, accessor_wrapper, **kwargs):
#     if isinstance(names, str):
#         names = [names]

#     def decorator(cls):
#         for name in names:
#             if hasattr(cls, name):
#                 warnings.warn(
#                     "registration of name {!r} for type {!r} is "
#                     "overriding a preexisting attribute with the same name.".format(
#                         name, cls
#                     ),
#                     AccessorRegistrationWarning,
#                     stacklevel=2,
#                 )
#             setattr(cls, name, accessor_wrapper(name, **kwargs))
#         return cls

#     return decorator


# def decorate_listaccessor(names, cache_list=None):
#     return _decorate_listaccessor(
#         names, _CachedListAccessorWrapper, cache_list=cache_list
#     )


# def decorate_listproperty(names, cache_list=None):
#     return _decorate_listaccessor(
#         names, _CachedListPropertyWrapper, cache_list=cache_list
#     )



# mypy: disable-error-code="no-untyped-def, no-untyped-call"
import numpy as np
import pandas as pd
import pytest

import lnpy


@pytest.fixture(scope="session")
def path_data(path_data_single_comp_super):
    return path_data_single_comp_super


@pytest.fixture(scope="session")
def ref(ref_single_comp_super):
    return ref_single_comp_super


@pytest.fixture(scope="session")
def phase_creator(ref):
    return lnpy.segment.PhaseCreator(
        nmax=1,
        nmax_peak=4,
        ref=ref,
        merge_kws={"efac": 0.8},
        # if want to id phases based on some
        # callable, then set this.
        # the callable is passes a list of
        # MaskedlnPi objects, and should return
        # a list of phaseid
        tag_phases=None,
        # tag_phases=tag_phases2
    )


@pytest.fixture(scope="session")
def build_phases(phase_creator):
    # choose a particular factory method for creating phases
    return phase_creator.build_phases_mu([None])


import lnpy.examples


# can drop param = 0 after make sure all good
@pytest.fixture(params=[1])
def obj(request, ref, phase_creator, build_phases):
    if request.param == 0:
        return lnpy.examples.Example(
            ref=ref, phase_creator=phase_creator, build_phases=build_phases
        )

    return lnpy.examples.lj_sup_example()


@pytest.fixture(scope="session")
def test_table(path_data):
    return pd.read_csv(path_data / "data_0.csv")


def get_test_table(o, ref):
    return (
        o.xge.table(
            keys=[
                "betaOmega",
                "nvec",
                "PE",
                "dens",
                "betaF",
                "S",
                "betaG",
                "edge_distance",
            ],
            ref=ref,
        )
        .to_dataframe()
        .reset_index()
    )  # .to_csv('data_0.csv', index=False)


def test_collection_properties(obj, test_table) -> None:
    ref, build_phases = obj.unpack(["ref", "build_phases"])

    # for big builds, take advantage of progress bar, and parallel builds
    lnzs = np.linspace(-10, 3.5, 2000)

    # by default, progress bar hides itself after completion.  use context manager to keep it
    # note that for this example (where only have a single phase), doesn't really make a difference
    with lnpy.set_options(tqdm_leave=True, joblib_use=False, tqdm_bar="text"):
        o = lnpy.lnPiCollection.from_builder(lnzs, build_phases)

    other = get_test_table(o, ref)

    pd.testing.assert_frame_equal(test_table, other)


@pytest.fixture(scope="session")
def test_table_can(path_data):
    return pd.read_csv(path_data / "data_0_can.csv")


def get_test_table_can(ref):
    return (
        ref.xce.table(
            keys=["S", "Z", "betaE", "betaF", "betaOmega", "betamu", "dens", "ntot"]
        )
        .to_dataframe()
        .reset_index()
    )


def test_canonical_properties(obj, test_table_can) -> None:
    ref = obj.ref
    other = get_test_table_can(ref)
    pd.testing.assert_frame_equal(other, test_table_can)


def test_nice_grid(obj, path_data) -> None:
    ref, build_phases = obj.unpack(["ref", "build_phases"])
    import lnpy.lnpicollectionutils

    with lnpy.set_options(joblib_use=True):
        o_course, o = lnpy.lnpicollectionutils.limited_collection(
            build_phases,
            dlnz=0.01,
            offsets=[-10, +10],
            even_grid=True,  # but lnzs on same grid as dlnz
            digits=2,  # round lnzs to this number of digits
            edge_distance_min=10,
            dens_min=0.001,
        )

    other_course = get_test_table(o_course, ref)
    test_course = pd.read_csv(path_data / "data_0_course.csv")
    pd.testing.assert_frame_equal(other_course, test_course)

    other_fine = get_test_table(o, ref)
    test = pd.read_csv(path_data / "data_0_fine.csv")
    pd.testing.assert_frame_equal(other_fine, test)



"""routines to find constant molfracs"""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

from .lnpiseries import lnPiCollection
from .utils import RootResultDict, array_to_scalar, rootresults_to_rootresultdict

if TYPE_CHECKING:
    from typing import Any, Mapping

    import xarray as xr

    from .lnpidata import lnPiMasked
    from .segment import BuildPhasesBase


def _initial_bracket_molfrac(
    target: float,
    collection: lnPiCollection,
    build_phases: BuildPhasesBase,
    phase_id: int | str | None = 0,
    component: int | None = None,
    dlnz: float = 0.5,
    dfac: float = 1.0,
    ref: lnPiMasked | None = None,
    build_kws: Mapping[str, Any] | None = None,
    ntry: int = 20,
) -> tuple[lnPiCollection, lnPiCollection, dict[str, Any]]:
    """Find bracket for molfrac"""

    lnz_idx = build_phases.index
    if component is None:
        component = lnz_idx
    if build_kws is None:
        build_kws = {}

    if isinstance(phase_id, str) and phase_id.lower() == "none":
        # select stable phase
        skip_phase_id = True

        def getter(x: lnPiCollection) -> xr.DataArray:
            v = x.xge.molfrac.sel(component=component)
            if len(x) == 1:
                return v.isel(phase=0)
            return v.where(x.xge.mask_stable).max("phase")

    else:
        skip_phase_id = False

        def getter(x: lnPiCollection) -> xr.DataArray:
            return x.xge.molfrac.sel(component=component, phase=phase_id)

    s = getter(collection).to_series().dropna()

    # get left bound
    left = None
    ntry_left = 0
    ss = s[s < target]
    if len(ss) > 0:
        left = collection.mloc[ss.index[[-1]]]
    else:
        index = s.index[[0]] if len(s) > 0 else collection.index[[0]].droplevel("phase")
        new_lnz = collection.mloc[index]._get_lnz(lnz_idx)
        dlnz_ = dlnz
        for i in range(ntry):  # noqa: B007
            new_lnz -= dlnz_
            dlnz_ *= dfac
            p = build_phases(new_lnz, ref=ref, **build_kws)
            if (skip_phase_id or phase_id in p._get_level("phase")) and getter(
                p
            ).to_numpy() < target:
                left = p
                break
        ntry_left = i

    if left is None:
        msg = "could not find left bounds"
        raise RuntimeError(msg)

    # right bracket
    right = None
    ntry_right = 0
    ss = s[s > target]
    if len(ss) > 0:
        right = collection.mloc[ss.index[[0]]]
    else:
        index = (
            s.index[[-1]] if len(s) > 0 else collection.index[[-1]].droplevel("phase")
        )
        new_lnz = collection.mloc[index]._get_lnz(lnz_idx)
        dlnz_ = dlnz

        for i in range(ntry):  # noqa: B007
            new_lnz += dlnz_
            p = build_phases(new_lnz, ref=ref, **build_kws)
            if (not skip_phase_id) and (phase_id not in p._get_level("phase")):
                # went to far
                new_lnz -= dlnz_
                # reset to half dlnz
                dlnz_ = dlnz_ * 0.5
            elif getter(p).to_numpy() > target:
                right = p
                break
            else:
                dlnz_ *= dfac
        ntry_right = i

    if right is None:
        msg = "could not find right bounds"
        raise RuntimeError(msg)

    info = {"ntry_left": ntry_left, "ntry_right": ntry_right}
    return left, right, info


def _solve_lnz_molfrac(
    target: float,
    left: lnPiCollection | float,
    right: lnPiCollection | float,
    build_phases: BuildPhasesBase,
    phase_id: int | str | None = 0,
    component: int | None = None,
    build_kws: Mapping[str, Any] | None = None,
    ref: lnPiMasked | None = None,
    tol: float = 1e-4,
    **kwargs: Any,
) -> tuple[lnPiCollection, RootResultDict]:
    """
    Calculate `lnz` which provides ``molfrac == target``.

    Parameters
    ----------
    target : float
        target molfraction
    a, b : float
        lnz values bracketing solution
    phase_id : int
        target phase
    ref : lnPiMasked
        object to reweight
    component : int, optional
        if not specified, use build_phases.index
    full_output : bool, default=False
        if True, return solve stats
    tol : float, default=1e-4
        solver tolerance

    **kwargs : extra arguments to scipy.optimize.brentq

    Returns
    -------
    output : lnPi_phases object
        object with desired molfraction
    info : solver info (optional, returned if full_output is `True`)
    """
    from scipy.optimize import brentq

    if build_kws is None:
        build_kws = {}

    if component is None:
        component = build_phases.index

    if isinstance(left, lnPiCollection):
        left = left._get_lnz(build_phases.index)
    if isinstance(right, lnPiCollection):
        right = right._get_lnz(build_phases.index)

    a, b = sorted((left, right))

    if isinstance(phase_id, str) and phase_id.lower() == "none":
        # select stable phase
        skip_phase_id = True

        def getter(x: lnPiCollection) -> xr.DataArray:
            v = x.xge.molfrac.sel(component=component)
            if len(x) == 1:
                return v.isel(phase=0)
            return v.where(x.xge.mask_stable).max("phase")

    else:
        skip_phase_id = False

        def getter(x: lnPiCollection) -> xr.DataArray:
            return x.xge.molfrac.sel(component=component, phase=phase_id)

    def f(x: float) -> float:
        p = build_phases(x, ref=ref, **build_kws)
        f.lnpi = p  # type: ignore[attr-defined]

        # by not using the ListAccessor,
        # can parallelize
        mf: float
        if skip_phase_id or phase_id in p._get_level("phase"):
            mf = array_to_scalar(getter(p).values)
        else:
            mf = np.inf

        return mf - target

    xx, r = brentq(f, a, b, full_output=True, **kwargs)

    residual = f(xx)
    if np.abs(residual) > tol:
        msg = "something went wrong with solve"
        raise RuntimeError(msg)

    return f.lnpi, rootresults_to_rootresultdict(r, residual=residual)  # type: ignore[attr-defined]


def find_lnz_molfrac(
    target: float,
    collection: lnPiCollection,
    build_phases: BuildPhasesBase,
    phase_id: int | str | None = 0,
    component: int | None = None,
    build_kws: Mapping[str, Any] | None = None,
    ref: lnPiMasked | None = None,
    dlnz: float = 0.5,
    dfac: float = 1.0,
    ntry: int = 20,
    tol: float = 1e-4,
    **kwargs: Any,
) -> tuple[lnPiCollection, RootResultDict]:
    left, right, _info = _initial_bracket_molfrac(
        target=target,
        collection=collection,
        build_phases=build_phases,
        phase_id=phase_id,
        component=component,
        dlnz=dlnz,
        dfac=dfac,
        ref=ref,
        build_kws=build_kws,
        ntry=ntry,
    )

    lnpi, r = _solve_lnz_molfrac(
        target=target,
        left=left,
        right=right,
        build_phases=build_phases,
        phase_id=phase_id,
        component=component,
        build_kws=build_kws,
        ref=ref,
        tol=tol,
        **kwargs,
    )
    return lnpi, r



# mypy: disable-error-code="no-untyped-def, no-untyped-call"
from __future__ import annotations

import json
import locale
from pathlib import Path

import numpy as np
import pandas as pd
import pytest

from lnpy import lnPiMasked


def get_lnz(path):
    k_b = 1.3806503e-23  # J/K
    kg = 1.66054e-27  # 1/amu
    h_planck = 6.62606876e-34  # Js

    with Path(path).open(encoding=locale.getpreferredencoding(False)) as f:
        metadata = json.load(f)

    debroglie = (
        h_planck
        / np.sqrt(
            2.0
            * np.pi
            * (metadata["mass"] * kg)
            * k_b
            * (metadata["T*"] * metadata["eps_kB"])
        )
        * 1.0e10
    )  # ang

    mu = metadata["mu*"] + 3.0 * metadata["T*"] * np.log(metadata["sigma"] / debroglie)
    temp = metadata["T*"]

    lnz = mu / temp

    return lnz, {"beta": 1.0 / temp, "volume": metadata["V*"]}


def get_ref(
    path_data: str | Path,
    metadata_name: str,
    lnpi_name: str,
    energy_name: str,
) -> lnPiMasked:
    path_data = Path(path_data)
    lnz, state_kws = get_lnz(path_data / metadata_name)

    # read in potential energy
    pe = pd.read_csv(
        path_data / energy_name,
        header=None,
        sep="\\s+",
        names=["n", "e"],
    )["e"].to_numpy()

    return (
        lnPiMasked.from_table(
            path_data / lnpi_name,
            fill_value=np.nan,
            lnz=lnz,
            # state_kws needs to be defined if want to calculate some properties
            # down the road
            state_kws=state_kws,
            # extra_kws is where you pass things which will be passed along to
            # new lnPis, like potential energy
            extra_kws={"PE": pe},
        )
        .zeromax()
        .pad()
    )


@pytest.fixture(scope="session")
def path_data_lnPi():  # noqa: N802
    return Path(__file__).parent / "../examples/archived/LJ_cfs_2.5sig"


@pytest.fixture(scope="session")
def ref_lnPi(path_data_lnPi):  # noqa: N802,N803
    import lnPi

    path_data = path_data_lnPi

    lnz, state_kws = get_lnz(path_data / "t150.metadata.json")

    # read in potential energy
    pe = pd.read_csv(
        path_data / "ljsf.t150.bulk.v512.r1.energy.dat",
        header=None,
        sep="\\s+",
        names=["n", "e"],
    )["e"].to_numpy()

    return (
        lnPi.MaskedlnPiDelayed.from_table(
            path_data / "ljsf.t150.bulk.v512.r1.lnpi.dat",
            fill_value=np.nan,
            lnz=lnz,
            # state_kws needs to be defined if want to calculate some properties
            # down the road
            state_kws=state_kws,
            # extra_kws is where you pass things which will be passed along to
            # new lnPis, like potential energy
            extra_kws={"PE": pe},
        )
        .zeromax()
        .pad()
    )


@pytest.fixture(scope="session")
def path_data_single_comp_super():
    return Path(__file__).parent / "../examples/archived/LJ_cfs_2.5sig"


@pytest.fixture(scope="session")
def ref_single_comp_super(path_data_single_comp_super):
    return get_ref(
        path_data=path_data_single_comp_super,
        metadata_name="t150.metadata.json",
        energy_name="ljsf.t150.bulk.v512.r1.energy.dat",
        lnpi_name="ljsf.t150.bulk.v512.r1.lnpi.dat",
    )


@pytest.fixture(scope="session")
def path_data_single_comp_sub():
    return Path(__file__).parent / "../examples/archived/LJ_cfs_2.5sig"


@pytest.fixture(scope="session")
def ref_single_comp_sub(path_data_single_comp_super):
    return get_ref(
        path_data=path_data_single_comp_super,
        metadata_name="t072871.metadata.json",
        energy_name="ljsf.t072871.bulk.v512.r1.energy.dat",
        lnpi_name="ljsf.t072871.bulk.v512.r1.lnpi.dat",
    )



"""
A script to invoke utilities from either pipx or from local installs.

I got tired of installing mypy/pyright in a million venvs. I got around this
for pyright using a central install and running in an activated venv. You can
do likewise with mypy, by passing `--python-executable`. I got the bright idea
to use `pipx` to manage mypy and pyright. But when working from my own machine,
I'd already have the type checkers installed centrally. This script automates
running these tools. It does the following:

* Optionally can set the `specification` (i.e., "mypy==1.2.3...", etc)
* Will check if the specification is installed. If it is, use it (unless pass
 `-x`). Otherwise, run the command via `pipx` (something like `pipx run
 mypy==1.2.3...`)
* You can set the specifications from a `requirements.txt` file. So you can use
  tools like `pip-compile` to manage the versions.
* Makes the `--python-executable` and `--pythonpath` flags to mypy and pyright
  the same. Defaults to using `sys.executable` from the python running this
* Also sets `--python-version` and `--pythonversion` in mypy and pyright
* For other tools, just run them from pipx or installed.
"""

from __future__ import annotations

import logging
import os
import re
import shlex
import subprocess
import sys
from functools import lru_cache
from pathlib import Path
from typing import TYPE_CHECKING, Any, Protocol, cast

from packaging.requirements import Requirement
from packaging.version import Version

if TYPE_CHECKING:
    import argparse
    from types import ModuleType
    from typing import Iterable, Iterator, Mapping, Sequence

# Optional imports
colorlog: ModuleType | None
try:
    import colorlog
except ImportError:
    colorlog = None


# * Logger --------------------------------------------------------------------
# Taken from nox.logger (https://github.com/wntrblm/nox/tree/main/nox/logger.py)

SUCCESS = 25
OUTPUT = logging.DEBUG - 1


def _get_format(color: bool, add_timestamp: bool) -> str:
    if color and colorlog:
        if add_timestamp:
            return "%(cyan)s%(name)s > [%(asctime)s] %(log_color)s%(message)s"
        return "%(cyan)s%(name)s > %(log_color)s%(message)s"

    if add_timestamp:
        return "%(name)s > [%(asctime)s] %(message)s"

    return "%(name)s > %(message)s"


class PipxRunFormatter(logging.Formatter):
    """Custom formatter."""

    def __init__(self, add_timestamp: bool = False) -> None:
        super().__init__(fmt=_get_format(color=False, add_timestamp=add_timestamp))
        self._simple_fmt = logging.Formatter("%(message)s")

    def format(self, record: Any) -> str:
        if record.levelname == "OUTPUT":
            return self._simple_fmt.format(record)
        return super().format(record)


if colorlog:

    class PipxRunColoredFormatter(colorlog.ColoredFormatter):  # type: ignore[misc,name-defined]
        """Colored formatter."""

        def __init__(
            self,
            datefmt: Any = None,
            style: Any = None,
            log_colors: Any = None,
            reset: bool = True,
            secondary_log_colors: Any = None,
            add_timestamp: bool = False,
        ) -> None:
            super().__init__(  # pyright: ignore[reportUnknownVariableType, reportUnknownMemberType]
                fmt=_get_format(color=True, add_timestamp=add_timestamp),
                datefmt=datefmt,
                style=style,
                log_colors=log_colors,
                reset=reset,
                secondary_log_colors=secondary_log_colors,
            )
            self._simple_fmt = logging.Formatter("%(message)s")

        def format(self, record: Any) -> str:
            if record.levelname == "OUTPUT":
                return self._simple_fmt.format(record)
            return super().format(record)  # type: ignore[no-any-return]


class LoggerWithSuccessAndOutput(logging.getLoggerClass()):  # type: ignore[misc]
    """Custom Logger."""

    def __init__(self, name: str, level: int = logging.NOTSET) -> None:
        super().__init__(name, level)
        logging.addLevelName(SUCCESS, "SUCCESS")
        logging.addLevelName(OUTPUT, "OUTPUT")

    def success(self, msg: str, *args: Any, **kwargs: Any) -> None:
        if self.isEnabledFor(SUCCESS):  # pragma: no cover
            self._log(SUCCESS, msg, args, **kwargs)

    def output(self, msg: str, *args: Any, **kwargs: Any) -> None:
        if self.isEnabledFor(OUTPUT):  # pragma: no cover
            self._log(OUTPUT, msg, args, **kwargs)


logging.setLoggerClass(LoggerWithSuccessAndOutput)
logger = cast(LoggerWithSuccessAndOutput, logging.getLogger("pipxrun"))


def _get_formatter(color: bool, add_timestamp: bool) -> logging.Formatter:
    if color and colorlog:
        return PipxRunColoredFormatter(
            reset=True,
            log_colors={
                "DEBUG": "cyan",
                "INFO": "blue",
                "WARNING": "yellow",
                "ERROR": "red",
                "CRITICAL": "red,bg_white",
                "SUCCESS": "green",
            },
            style="%",
            secondary_log_colors=None,
            add_timestamp=add_timestamp,
        )
    return PipxRunFormatter(add_timestamp=add_timestamp)


def setup_logging(
    color: bool, verbosity: int = 0, add_timestamp: bool = False
) -> None:  # pragma: no cover
    """Setup logging.

    Args:
        color (bool): If true, the output will be colored using
            colorlog. Otherwise, it will be plaintext.
    """
    root_logger = logging.getLogger()

    level_number = max(0, logging.WARNING - 10 * verbosity)
    root_logger.setLevel(level_number)

    # if verbose:
    #     root_logger.setLevel(OUTPUT)
    # else:
    #     root_logger.setLevel(logging.DEBUG)
    handler = logging.StreamHandler()

    handler.setFormatter(_get_formatter(color, add_timestamp))
    root_logger.addHandler(handler)

    # Silence noisy loggers
    logging.getLogger("sh").setLevel(logging.WARNING)


# Original
# class CustomFormatter(logging.Formatter):
#     """Custom formatter."""

#     def format(self, record: logging.LogRecord) -> str:
#         """Custom format."""
#         msg = textwrap.fill(
#             super().format(record),
#             width=_max_output_width(),
#             subsequent_indent=" " * 4,
#             break_long_words=False,
#             replace_whitespace=False,
#             drop_whitespace=False,
#             break_on_hyphens=False,
#         ).replace("\n", "\\\n")
#         if msg.endswith("\\\n"):
#             msg = msg.rstrip("\\\n") + "\n"

#         return msg


# handler = logging.StreamHandler(sys.stdout)
# handler.setFormatter(
#     CustomFormatter(
#         "{name}({levelname}): {message}",
#         style="{",
#     )
# )

# logging.basicConfig(
#     level=logging.WARNING,
#     handlers=[
#         handler,
#     ],
# )
# logger = logging.getLogger("pipxrun")

# def _set_verbosity_level(logger: logging.Logger, verbosity: int) -> None:
#     """Set verbosity level."""
#     level_number = max(0, logging.WARNING - 10 * verbosity)
#     logger.setLevel(level_number)


# * Utilities -----------------------------------------------------------------
@lru_cache
def _comment_re() -> re.Pattern[str]:
    return re.compile(r"(^|\s+)#.*$")


# @lru_cache
# def _max_output_width() -> int:
#     import shutil

#     width = shutil.get_terminal_size((80, 20)).columns

#     min_width, max_width = 20, 150
#     if width > max_width:
#         width = max_width
#     if width < min_width:
#         width = min_width
#     return width


# def _print_header(name: str = "") -> None:
#     if logger.isEnabledFor(logging.WARNING):
#         fmt = f"{{:=<{_max_output_width()}}}"
#         if name:
#             name = f"= {name} "
#         print(fmt.format(name))


@lru_cache
def _get_command_version(name: str, path: str) -> Version:
    # Note that this version is faster than
    # Calling subprocess to get --version (see bottom of file).
    # But it might break on windows...

    with Path(path).open() as f:
        python_executable = f.readline().strip().replace("#!", "")

    return Version(
        subprocess.check_output(
            [
                python_executable,
                "-c",
                f"from importlib.metadata import version; print(version('{name}'))",
            ]
        )
        .decode()
        .strip()
    )


def _parse_requirements(requirements: Path) -> Iterator[Requirement]:
    comment_re = _comment_re()

    def ignore_comments(lines: Iterable[str]) -> Iterator[str]:
        """Strips comments and filter empty lines."""
        for line in lines:
            if line_formatted := comment_re.sub("", line).strip():
                yield line_formatted

    with requirements.open() as f:
        yield from map(Requirement, ignore_comments(f))


@lru_cache
def _get_python_version(python_version: str | None) -> str:
    if python_version is None:
        return "{}.{}".format(*sys.version_info[:2])
    return python_version


@lru_cache
def _get_python_executable(python_executable: str | None) -> str:
    if python_executable is None:
        return sys.executable
    return python_executable


def _iter_specs(specs: Iterable[str | Requirement]) -> Iterator[Requirement]:
    for spec in specs:
        if isinstance(spec, str):
            yield Requirement(spec)
        else:
            yield spec


# * Specs ---------------------------------------------------------------------
class Specifications:
    """Working with specifications."""

    def __init__(self, specs: dict[str, Requirement]) -> None:
        self.specs = specs

    def assign(
        self, specs: str | Requirement | Iterable[str | Requirement]
    ) -> Specifications:
        if isinstance(specs, (str, Requirement)):
            specs = [specs]

        return type(self)(
            specs=dict(self.specs, **{spec.name: spec for spec in _iter_specs(specs)})
        )

    def get(self, index: str, default: Requirement | None = None) -> Requirement | None:
        return self.specs.get(index, default)

    @classmethod
    def combine(cls, *specs: Specifications) -> Specifications:
        new_specs: dict[str, Requirement] = {}
        for spec in specs:
            new_specs.update(spec.specs)
        return cls(new_specs)

    @classmethod
    def from_requirements(
        cls,
        *specs: Requirement,
        requirements: str | Path | Iterable[str | Path] | None = None,
    ) -> Specifications:
        if requirements is None:
            requirements = []
        elif isinstance(requirements, (str, Path)):
            requirements = [requirements]

        specs_dict: dict[str, Requirement] = {
            req.name: req
            for requirement in requirements
            for req in _parse_requirements(Path(requirement))
        }

        # update specs
        specs_dict.update({spec.name: spec for spec in _iter_specs(specs)})

        return cls(specs_dict)


# * Dummy session -------------------------------------------------------------
class SessionInterfaceTemplate(Protocol):
    """Generic session."""

    def run(
        self,
        *args: str | os.PathLike[str],
        env: Mapping[str, str] | None,
        include_outer_env: bool = True,
        **kwargs: Any,
    ) -> Any | None: ...

    def log(self, *args: Any, **kwargs: Any) -> None: ...

    def warn(self, *args: Any, **kwargs: Any) -> None: ...

    def debug(self, *args: Any, **kwargs: Any) -> None: ...


class Session:
    """Basic session"""

    def log(self, *args: Any, **kwargs: Any) -> None:
        """Alias to self.info"""
        self.info(*args, **kwargs)

    @staticmethod
    def info(*args: Any, **kwargs: Any) -> None:
        """Output log info."""
        logger.info(*args, **kwargs)

    @staticmethod
    def warn(*args: Any, **kwargs: Any) -> None:
        """Output log.warning"""
        logger.warning(*args, **kwargs)

    @staticmethod
    def debug(*args: Any, **kwargs: Any) -> None:
        """Debug."""
        logger.debug(*args, **kwargs)

    def run(
        self,
        *args: str | os.PathLike[str],
        env: Mapping[str, str] | None = None,
        include_outer_env: bool = True,  # noqa: ARG002
        **kwargs: Any,  # noqa: ARG002
    ) -> None:
        """Actually do the run."""
        cleaned_args = [os.fsdecode(arg) for arg in args]

        full_cmd = shlex.join(cleaned_args)

        self.info("Running %s", full_cmd)
        r = subprocess.run(cleaned_args, check=False, env=env)  # pyright: ignore[reportUnknownVariableType]

        returncode: int = r.returncode  # pyright: ignore[reportUnknownVariableType, reportUnknownMemberType]
        if returncode != 0:
            logger.error("Command %s failed with exit code %s", full_cmd, returncode)  # pyright: ignore[reportUnknownArgumentType]
            msg = f"Returned code {returncode}"
            raise RuntimeError(msg)


# * Command -------------------------------------------------------------------
class Command:
    """Class to handle commands."""

    def __init__(
        self, name: str, args: Iterable[str], spec: Requirement | None = None
    ) -> None:
        self.name = name
        self.args = list(args)
        self.spec = spec

    @classmethod
    def from_command(cls, *commands: str | os.PathLike[str]) -> Command:
        """Create from command iterable"""
        name, *args = map(os.fsdecode, commands)
        spec = Requirement(name)
        return cls(
            name=spec.name,
            args=args,
            spec=spec if spec.specifier else None,
        )

    def assign_spec(self, spec: Requirement | None, override: bool = False) -> Command:
        """Update spec from specs dict."""
        if spec and (override or not self.spec):
            return type(self)(name=self.name, args=self.args, spec=spec)
        return self

    def _get_python_flags(
        self,
        python_executable: str | None,
        python_version: str | None,
    ) -> list[str]:
        if self.name == "mypy":
            return [
                f"--python-executable={_get_python_executable(python_executable)}",
                f"--python-version={_get_python_version(python_version)}",
            ]
        if self.name == "pyright":
            return [
                f"--pythonpath={_get_python_executable(python_executable)}",
                f"--pythonversion={_get_python_version(python_version)}",
            ]

        return []

    @staticmethod
    def _get_pipx_flags(verbosity: int) -> list[str]:
        if verbosity > 0:
            return [f"-{'v' * verbosity}"]
        if verbosity < 0:
            return [f"-{'q' * -verbosity}"]
        return []

    def command(
        self,
        session: SessionInterfaceTemplate,
        python_executable: str | None,
        python_version: str | None,
        files: str | Iterable[str] | None,
        pipx_only: bool = False,
        verbosity: int = 0,
    ) -> list[str]:
        """Create command list."""
        from shutil import which

        if files is None:
            files = []
        elif isinstance(files, str):
            files = [files]

        commands: list[str] = []
        if not pipx_only and (exe_path := which(self.name)):
            if self.spec is None:
                session.log("Using local %s at %s", self.name, exe_path)
                commands = [exe_path]
            elif (
                exe_version := _get_command_version(name=self.name, path=exe_path)
            ) and exe_version in self.spec.specifier:
                session.log(
                    "Using local %s with version %s at %s",
                    self.spec,
                    exe_version,
                    exe_path,
                )
                commands = [exe_path]

        if not commands:
            session.warn("Using pipx run %s", self.spec or self.name)
            commands = [
                "pipx",
                "run",
                *self._get_pipx_flags(verbosity=verbosity),
                *([f"--spec={self.spec}"] if self.spec else []),
                self.name,
            ]

        return [
            *commands,
            *self._get_python_flags(
                python_executable=python_executable,
                python_version=python_version,
            ),
            *self.args,
            *files,
        ]

    def run(
        self,
        session: SessionInterfaceTemplate,
        python_executable: str | None,
        python_version: str | None,
        files: Iterable[str] | None = None,
        dry: bool = False,
        verbosity: int = 0,
        pipx_only: bool = False,
        env: Mapping[str, str] | None = None,
        include_outer_env: bool = True,
        **kwargs: Any,
    ) -> Any:
        """Run the command"""
        # _print_header(self.name)

        command = self.command(
            session=session,
            python_executable=python_executable,
            python_version=python_version,
            files=files,
            pipx_only=pipx_only,
            verbosity=verbosity,
        )

        if dry:
            session.log("Would run with commands: %s", command)
            return 0

        return session.run(
            *command, env=env, include_outer_env=include_outer_env, **kwargs
        )


def run(
    *args: str | os.PathLike[str],
    session: SessionInterfaceTemplate | None = None,
    specs: Specifications | None = None,
    extra_specs: str | Requirement | Iterable[str | Requirement] | None = None,
    python_version: str | None = None,
    python_executable: str | None = None,
    verbosity: int = -2,
    pipx_only: bool = False,
    env: Mapping[str, str] | None = None,
    files: Iterable[str] | None = None,
    include_outer_env: bool = True,
    **kwargs: Any,
) -> Any:
    """Run command"""
    command = Command.from_command(*args)

    if specs is None:
        specs = Specifications({})

    if extra_specs is not None:
        specs = specs.assign(extra_specs)

    command = command.assign_spec(specs.get(command.name))

    session = session or Session()

    return command.run(
        session=session,
        python_executable=python_executable,
        python_version=python_version,
        files=files,
        verbosity=verbosity,
        pipx_only=pipx_only,
        env=env,
        include_outer_env=include_outer_env,
        **kwargs,
    )


# * CLI -----------------------------------------------------------------------
def _parse_args(args: Sequence[str] | None = None) -> argparse.Namespace:
    """Get parser."""
    from argparse import ArgumentParser

    parser = ArgumentParser(
        description="Run executable using pipx or from installed if matches specification."
    )
    parser.add_argument(
        "--python-executable",
        dest="python_executable",
        default=None,  # Path(sys.executable),
        type=Path,
        help="""
        Path to python executable. Defaults to ``sys.executable``. This is
        passed to `--python-executable` in mypy and `--pythonpath` in pyright.
        """,
    )
    parser.add_argument(
        "--python-version",
        dest="python_version",
        default=None,
        type=str,
        help="""
        Python version (x.y) to typecheck against. Defaults to
        ``{sys.version_info.major}.{sys.version_info.minor}``. This is passed
        to ``--python-version`` and ``--pythonversion`` in mypy and pyright.
        """,
    )
    parser.add_argument(
        "-c",
        "--command",
        dest="commands",
        default=[],
        action="append",
        type=str,
        help="""
        Checkers command. This can include extra flags to the checker. This can
        also be passed multiple times for multiple checkers. This can also
        include a Requirement specification, which overrides any specification
        from ``--requirement`` or ``--spec``. For example,
        ``--command='mypy==1.8.0 --no-incremental' -c pyright``
        """,
    )
    parser.add_argument(
        "-r",
        "--requirement",
        dest="requirements",
        default=[],
        action="append",
        type=Path,
        help="Requirements (requirements.txt) specs for checker.  Can specify multiple times.",
    )
    parser.add_argument(
        "-s",
        "--spec",
        dest="specs",
        default=[],
        action="append",
        type=Requirement,
        help="""
        Package specification. Can pass multiple times. Overrides specs read
        from ``--requirements``. For example, ``--spec 'mypy==1.2.3'``.
        """,
    )
    parser.add_argument("--dry", action="store_true", help="Do dry run")
    parser.add_argument(
        "-v",
        "--verbose",
        dest="verbosity",
        action="count",
        default=0,
        help="Set verbosity level.  Pass multiple times to up level.",
    )
    parser.add_argument(
        "-q",
        "--quiet",
        action="count",
        default=0,
        help="Lower verbosity level.  Pass multiple times to up level.",
    )
    parser.add_argument(
        "-x",
        "--pipx",
        dest="pipx_only",
        action="store_true",
        help="Always use pipx to run the commands (No local fallback)",
    )
    parser.add_argument(
        "--fail-fast",
        dest="fail_fast",
        action="store_true",
        help="If passed, exit at first error.",
    )
    parser.add_argument(
        "files", type=str, default=[], nargs="*", help="Optional files to check."
    )

    options = parser.parse_args() if args is None else parser.parse_args(args)

    if not options.commands:
        parser.print_usage()

    return options


def main(args: Sequence[str] | None = None) -> int:
    """Main script."""
    if not (options := _parse_args(args)).commands:
        return 0

    # setup logging:
    setup_logging(
        color=True, verbosity=options.verbosity - options.quiet, add_timestamp=False
    )

    logger.info(
        "Running with python %s at %s",
        _get_python_version(None),
        _get_python_executable(None),
    )

    commands = [
        Command.from_command(*shlex.split(command)) for command in options.commands
    ]

    # specs from requirements
    specs = Specifications.from_requirements(
        *options.specs, requirements=options.requirements
    )

    # update command specs
    commands = [command.assign_spec(specs.get(command.name)) for command in commands]

    session = Session()

    s = 0

    for command in commands:
        try:
            command.run(
                session=session,
                python_executable=options.python_executable,
                python_version=options.python_version,
                dry=options.dry,
                verbosity=options.verbosity - options.quiet - 2,
                pipx_only=options.pipx_only,
                files=options.files,
            )
        except RuntimeError:  # noqa: PERF203
            s += 1
            if options.fail_fast:
                return s
    return s


if __name__ == "__main__":
    sys.exit(main())


# Alternative method to get command version...
# @lru_cache
# def _version_re() -> re.Pattern[str]:
#     return re.compile(
#         r"\b([1-9][0-9]*!)?(0|[1-9][0-9]*)"
#         r"(\.(0|[1-9][0-9]*))*((a|b|rc)(0|[1-9][0-9]*))"
#         r"?(\.post(0|[1-9][0-9]*))?(\.dev(0|[1-9][0-9]*))?\b"
#     )
#
#
# @lru_cache
# def _get_command_version(command: str, flag: str = "--version") -> Version | None:
#     version_search = _version_re().search(
#         subprocess.check_output([command, flag])
#         .decode()
#         .strip()
#     )
#     if version_search:
#         return Version(version_search.group(0))
#     return None
#
#
# def _get_site(python_executable: str | Path) -> str:
#     """Getesite-packages for python path (trying to make pytype work)"""
#     return (
#         subprocess.check_output(
#             [
#                 python_executable,
#                 "-c",
#                 "import sysconfig; print(sysconfig.get_paths()['purelib'])",
#             ]
#         )
#         .decode()
#         .strip()
#     )



# pyright: reportUnknownMemberType=false, reportUnknownArgumentType=false, reportUnknownVariableType=false
r"""
Routines to combine :math:`\ln \Pi` data (:mod:`~lnpy.combine`)
===============================================================

"""

from __future__ import annotations

from typing import TYPE_CHECKING, Iterator, cast

import numpy as np
import pandas as pd
from scipy.sparse import coo_array

if TYPE_CHECKING:
    from typing import Any, Iterable

    from numpy.typing import NDArray


class OverlapError(ValueError):
    """Specific error for missing overlaps."""


def _str_or_iterable(x: str | Iterable[str]) -> list[str]:
    if isinstance(x, str):
        return [x]
    return list(x)


# * connected graph
# From networkx (see https://github.com/networkx/networkx/blob/main/networkx/algorithms/components/connected.py)
def _plain_bfs(adj: dict[int, set[int]], source: int) -> set[int]:
    """A fast BFS node generator"""
    n = len(adj)
    seen = {source}
    nextlevel = [source]
    while nextlevel:
        thislevel = nextlevel
        nextlevel = []
        for v in thislevel:
            for w in adj[v]:
                if w not in seen:
                    seen.add(w)
                    nextlevel.append(w)
            if len(seen) == n:
                return seen
    return seen


def _connected_components(adj: dict[int, set[int]]) -> Iterator[set[int]]:
    seen = set()
    for v in adj:
        if v not in seen:
            c = _plain_bfs(adj, v)
            seen.update(c)
            yield c


def _build_graph(nodes: Iterable[int], edges: NDArray[Any]) -> dict[int, set[int]]:
    graph: dict[int, set[int]] = {node: set() for node in nodes}
    for a, b in edges:
        graph[a].add(b)
        graph[b].add(a)
    return graph


def check_windows_overlap(
    overlap_table: pd.DataFrame,
    windows: Iterable[int],
    state_names: str | Iterable[str] = "state",
    window_name: str = "window",
    verbose: bool = True,
) -> None:
    """
    Check that window overlaps form a connected graph.

    Parameters
    ----------
    overlap_table: DataFrame
        Frame should contain columns ``state_names`` and ``window_name``
        and rows corresponding to overlaps.

    Raises
    ------
    OverlapError
        If the overlaps do not form a connected graph, then raise a ``OverlapError``.
    """
    state_names = _str_or_iterable(state_names)
    overlap_table = overlap_table[[window_name, *state_names]]

    x: pd.DataFrame = (
        overlap_table.merge(
            overlap_table, on=state_names, how="outer", suffixes=("", "_nebr")
        )
        .drop(state_names, axis=1)
        .query(f"{window_name} < {window_name}_nebr")
        .drop_duplicates()
    )

    graph = _build_graph(nodes=windows, edges=x.to_numpy())

    components = list(_connected_components(graph))

    if len(components) != 1:
        msg = "Disconnected graph."
        if verbose:
            for subgraph in components:
                msg = f"{msg}\ngraph: {subgraph}"
        raise OverlapError(msg)


# * Matrix Equation construction
def _create_overlap_table(
    table: pd.DataFrame,
    window_name: str,
    window_max: int,
    state_names: list[str],
    lnpi_name: str,
    check_connected: bool = True,
) -> pd.DataFrame:
    overlap_table = table.loc[
        table[state_names].duplicated(keep=False),
        [window_name, *state_names, lnpi_name],
    ]

    if len(overlap_table) == 0:
        msg = "No overlaps with multiple windows"
        raise OverlapError(msg)

    if check_connected:
        check_windows_overlap(
            windows=range(window_max + 1),
            overlap_table=overlap_table,
            window_name=window_name,
            state_names=state_names,
        )
    return overlap_table


def _create_overlap_total_table(
    overlap_table: pd.DataFrame,
    window_name: str,
    window_max: int,
    state_names: list[str],
    lnpi_name: str,
) -> pd.DataFrame:
    group = overlap_table.groupby(state_names, as_index=False)[lnpi_name]
    return (
        overlap_table.assign(
            total=group.transform("sum"),
            count=group.transform("count"),
        )
        # limit to only window < window_max.
        # Equivalent to not shifting last window.
        .query(f"{window_name} < {window_max}")
        # assign eq_idx
        .assign(eq_idx=lambda x: range(len(x)))
    )


def _create_overlap_outer_table(
    overlap_total_table: pd.DataFrame,
    state_names: list[str],
    window_name: str,
) -> pd.DataFrame:
    return overlap_total_table[[*state_names, "eq_idx"]].merge(
        overlap_total_table[[*state_names, window_name]],
        on=state_names,
        how="outer",
    )


def _create_lhs_matrix_sparse(
    overlap_total_table: pd.DataFrame,
    overlap_outer_table: pd.DataFrame,
    window_name: str,
    window_max: int,
) -> coo_array:
    coeff_table = pd.concat(
        (
            overlap_total_table[["eq_idx", window_name, "count"]],
            overlap_outer_table[["eq_idx", window_name]].assign(count=-1),
        )
    )

    return coo_array(
        (
            coeff_table["count"],
            (coeff_table["eq_idx"], coeff_table[window_name]),
        ),
        shape=(len(overlap_total_table), window_max),
    )


def _create_lhs_matrix_numpy(
    overlap_total_table: pd.DataFrame,
    overlap_outer_table: pd.DataFrame,
    window_name: str,
    window_max: int,
) -> NDArray[Any]:
    overlap_outer_table = (
        overlap_outer_table[["eq_idx", window_name]]
        .assign(count=1)
        .groupby(["eq_idx", window_name], as_index=False)
        .sum()
    )
    a = np.zeros((len(overlap_total_table), window_max))
    a[overlap_total_table["eq_idx"], overlap_total_table["window"]] = (
        overlap_total_table["count"]
    )
    a[overlap_outer_table["eq_idx"], overlap_outer_table["window"]] -= (
        overlap_outer_table["count"]
    )
    return a

    # .. math::

    #     \min_{C_0, C_1, ..., C_{W-1}} \sum_{\rm{overlap}_m} \in \rm{overlaps}}\, \sum_{N_m, k \in \rm{overlap}_m} [\ln \bar{\Pi}(N_m) - (\ln \Pi_k (N_m) + C_k)]^2


def combine_lnpi(
    tables: Iterable[pd.DataFrame],
    state_names: str | Iterable[str] = "state",
    lnpi_name: str = "ln_prob",
    window_name: str = "window",
    use_sparse: bool = True,
    check_connected: bool = False,
) -> pd.DataFrame:
    r"""
    Merge multiple :math:`\ln \Pi` scans together in one pass using least squares.

    This performs least squares on the problem:

    .. math::

        \min_{C_0, C_1, ..., C_{W-1}} \sum_{\rm{overlap}_m \in \rm{overlaps}} \, \sum_{N_m, k \in \rm{overlap}_m}
        [\ln \bar{\Pi}(N_m) - (\ln \Pi_k (N_m) + C_k)]^2

    where,

    - :math:`C_j` : shift for sample :math:`j`.
    - :math:`W` : number of overlapping samples
    - :math:`\Pi_k(N)` : transition matrix at particle number :math:`N` for the kth sample
    - :math:`\rm{overlap}_m` : a particular overlap at particle number :math:`N_m` and over samples :math:`k`,
    - :math:`\ln \bar{\Pi}` is the to be determined average value.

    This can be reduced to a matrix problem of the form:

    .. math::

        S C_j - \sum_{k \in \rm{overlap}_m} C_k = - (S \ln \Pi_j(N_m)  - \sum_{k \in \rm{overlap}_m} \ln \Pi_k(N_m))

    the sum runs over all samples with overlap at state :math:`N_m`, :math:`S` is the number of such overlaps
    (i.e., :math:`S = \sum_{k \in \rm{overlap}_m} 1`).  There are such equations for all :math:`j \in \rm{overlap}_m`.



    Parameters
    ----------
    tables : iterable of DataFrame
        Individual sample windows.
    state_names : str or iterable of str
        Column name corresponding to a single state.  For example,
        for single component, this would be something like ``'n'``.  For binary system,
        this would be something like ``['n_0', 'n_1']``.
    lnpi_name : str
        Name of column in ``tables`` corresponding to :math:`\ln \Pi.`
    window_name : str, default="window"
        Name of column with will keep track of the window index.
    use_sparse : bool, default=True
        Use :class:`~scipy.sparse.coo_array` array in matrix equation. This is
        often faster than using a :class:`numpy.ndarray`.
    check_connected : bool, default=False
        If ``True``, check that all windows form a connected graph.

    Returns
    -------
    DataFrame
        DataFrame of all data.  The frame is not yet averaged over ``state_names``.
    """
    table: pd.DataFrame = pd.concat(dict(enumerate(tables)), names=[window_name, None])
    if window_name in table.columns:
        table = table.drop(window_name, axis=1)
    table = table.reset_index(window_name)

    window_max = cast(int, table[window_name].iloc[-1])
    if window_max == 0:
        return table

    state_names = _str_or_iterable(state_names)
    overlap_total_table = _create_overlap_total_table(
        overlap_table=_create_overlap_table(
            table,
            window_name=window_name,
            window_max=window_max,
            state_names=state_names,
            lnpi_name=lnpi_name,
            check_connected=check_connected,
        ),
        window_name=window_name,
        window_max=window_max,
        state_names=state_names,
        lnpi_name=lnpi_name,
    )

    b: NDArray[Any] = overlap_total_table.pipe(
        lambda x: -x["count"] * x[lnpi_name] + x["total"]
    ).to_numpy()

    overlap_outer_table = _create_overlap_outer_table(
        overlap_total_table=overlap_total_table,
        state_names=state_names,
        window_name=window_name,
    )

    if use_sparse:
        a = _create_lhs_matrix_sparse(
            overlap_total_table=overlap_total_table,
            overlap_outer_table=overlap_outer_table,
            window_name=window_name,
            window_max=window_max,
        )

        lhs = (a.T @ a).toarray()
        rhs = a.T @ b

    else:
        a = _create_lhs_matrix_numpy(
            overlap_total_table=overlap_total_table,
            overlap_outer_table=overlap_outer_table,
            window_name=window_name,
            window_max=window_max,
        )

        lhs = a.T @ a
        rhs = a.T @ b

    shift = np.zeros(window_max + 1)
    shift[:-1] = np.linalg.solve(lhs, rhs)
    shift -= shift[0]

    table[lnpi_name] += shift[table[window_name].to_numpy()]
    return table



"""
Collection of lnPi objects (:mod:`~lnpy.lnpiseries`)
====================================================
"""

from __future__ import annotations

import functools
from typing import TYPE_CHECKING, overload
from warnings import warn

import numpy as np
import pandas as pd
import xarray as xr
from module_utilities import cached

from .docstrings import docfiller
from .extensions import AccessorMixin

# lazy loads
from .utils import get_tqdm_build as get_tqdm
from .utils import labels_to_masks, masks_to_labels
from .utils import parallel_map_build as parallel_map

if TYPE_CHECKING:
    from typing import (
        Any,
        Callable,
        Final,
        Hashable,
        Iterable,
        Iterator,
        Literal,
        Mapping,
        Sequence,
    )

    from numpy.typing import ArrayLike, DTypeLike
    from pandas.core.groupby.generic import SeriesGroupBy

    from . import ensembles, lnpienergy, stability
    from ._typing import IndexingInt, MyNDArray, Scalar
    from ._typing_compat import Self
    from .lnpidata import lnPiMasked


# Accessors
class _CallableResult:
    def __init__(self, parent: lnPiCollection, func: Callable[..., Any]) -> None:
        functools.update_wrapper(self, func)

        self._parent = parent
        self._func = func

    def __call__(self, *args: Any, **kwargs: Any) -> lnPiCollection:
        return self._parent.new_like(self._func(*args, **kwargs))


class _Groupby:
    def __init__(self, parent: lnPiCollection, group: SeriesGroupBy[Any, Any]) -> None:
        self._parent = parent
        self._group = group

    def __iter__(self) -> Iterator[tuple[Any, lnPiCollection]]:
        return ((meta, self._parent.new_like(x)) for meta, x in self._group)

    def __getattr__(self, attr: str) -> _CallableResult | lnPiCollection:
        if hasattr(self._group, attr):
            out = getattr(self._group, attr)
            if callable(out):
                return _CallableResult(self._parent, out)
            return self._parent.new_like(out)

        msg = f"no attribute {attr} in groupby"
        raise AttributeError(msg)


# @SeriesWrapper.decorate_accessor("loc")
class _LocIndexer:
    """
    Indexer by value.

    See :attr:`pandas.Series.loc`
    """

    def __init__(self, parent: lnPiCollection) -> None:
        self._parent = parent
        self._loc = self._parent._series.loc

    @overload
    def __getitem__(self, idx: Scalar | tuple[Scalar, ...]) -> lnPiMasked: ...

    @overload
    def __getitem__(
        self,
        idx: list[Scalar] | pd.Index[Any] | slice | Callable[[pd.Series[Any]], Any],
    ) -> lnPiCollection: ...

    @overload
    def __getitem__(self, idx: Any) -> lnPiMasked | lnPiCollection: ...

    def __getitem__(self, idx: Any) -> lnPiMasked | lnPiCollection:
        out = self._loc[idx]
        if isinstance(out, pd.Series):
            out = self._parent.new_like(out)
        return out  # type: ignore[no-any-return]

    def __setitem__(
        self, idx: Any, values: lnPiMasked | pd.Series[Any] | Sequence[lnPiMasked]
    ) -> None:
        self._parent._series.loc[idx] = values  # pyright: ignore[reportCallIssue,reportArgumentType]


# @SeriesWrapper.decorate_accessor("iloc")
class _iLocIndexer:  # noqa: N801
    """
    Indexer by position.

    See :attr:`pandas.Series.iloc`
    """

    def __init__(self, parent: lnPiCollection) -> None:
        self._parent = parent
        self._iloc = self._parent._series.iloc

    @overload
    def __getitem__(self, idx: IndexingInt) -> lnPiMasked: ...

    @overload
    def __getitem__(
        self, idx: Sequence[int] | pd.Index[Any] | slice
    ) -> lnPiCollection: ...

    @overload
    def __getitem__(self, idx: Any) -> lnPiMasked | lnPiCollection: ...

    def __getitem__(self, idx: Any) -> lnPiMasked | lnPiCollection:
        out = self._iloc[idx]
        if isinstance(out, pd.Series):
            out = self._parent.new_like(out)
        return out  # type: ignore[no-any-return]

    def __setitem__(
        self, idx: Any, values: lnPiMasked | pd.Series[Any] | Sequence[lnPiMasked]
    ) -> None:
        self._parent._series.iloc[idx] = values  # pyright: ignore[reportCallIssue,reportArgumentType]


# @SeriesWrapper.decorate_accessor("query")
class _Query:
    """
    Select values by string query.

    See :meth:`pandas.DataFrame.query`
    """

    def __init__(self, parent: lnPiCollection) -> None:
        self._parent = parent
        self._frame: pd.DataFrame = self._parent.index.to_frame().reset_index(drop=True)

    def __call__(self, expr: str, **kwargs: Any) -> lnPiCollection:
        idx = self._frame.query(expr, **kwargs).index
        return self._parent.iloc[idx]  # type: ignore[no-any-return]


# @SeriesWrapper.decorate_accessor("zloc")
class _LocIndexer_unstack_zloc:  # noqa: N801
    """positional indexer for everything but phase"""

    def __init__(
        self,
        parent: lnPiCollection,
        level: Hashable | Sequence[Hashable] = ("phase",),
    ) -> None:
        self._parent = parent
        self._level = level
        self._loc = self._parent._series.unstack(self._level).iloc  # noqa: PD010

    def __getitem__(self, idx: Any) -> lnPiCollection:
        out = self._loc[idx]
        out = out.stack(self._level) if isinstance(out, pd.DataFrame) else out.dropna()  # noqa: PD013

        if isinstance(out, pd.Series):
            out = self._parent.new_like(out)
        else:
            msg = "unknown indexer for zloc"
            raise TypeError(msg)
        return out  # type: ignore[no-any-return]


# @SeriesWrapper.decorate_accessor("mloc")
class _LocIndexer_unstack_mloc:  # noqa: N801
    """indexer with pandas index"""

    def __init__(
        self,
        parent: lnPiCollection,
        level: Hashable | Sequence[Hashable] = ("phase",),
    ) -> None:
        self._parent = parent
        self._level = level
        self._index = self._parent.index

        self._index_names = set(self._index.names)
        self._loc = self._parent._series.iloc

    def _get_loc_idx(self, idx: pd.MultiIndex | pd.Index[Any]) -> Any:
        index = self._index
        if isinstance(idx, pd.MultiIndex):
            # names in idx and
            drop: list[Hashable] = list(self._index_names - set(idx.names))
            index = index.droplevel(drop)
            # reorder idx
            idx = idx.reorder_levels(index.names)  # type: ignore[no-untyped-call]
        else:
            drop = list(set(index.names) - {idx.name})
            index = index.droplevel(drop)
        return index.get_indexer_for(idx)  # type: ignore[no-untyped-call]

    def __getitem__(self, idx: pd.MultiIndex | pd.Index[Any]) -> lnPiCollection:
        indexer = self._get_loc_idx(idx)
        out = self._loc[indexer]

        if isinstance(out, pd.Series):
            out = self._parent.new_like(out)
        else:
            msg = "unknown indexer for mloc"
            raise TypeError(msg)
        return out  # type: ignore[no-any-return]


class lnPiCollection(AccessorMixin):  # noqa: PLR0904, N801
    r"""
    Wrapper around :class:`pandas.Series` for collection of :class:`~lnpy.lnpidata.lnPiMasked` objects.


    Parameters
    ----------
    data : sequence of lnPiMasked
        :math:`\ln \Pi(N)` instances to consider.
    index : array-like, pandas.Index, pandas.MultiIndex, optional
        Index to apply to Series.
    xarray_output : bool, default=True
        If True, then wrap lnPiCollection outputs in :class:`~xarray.DataArray`
    concat_dim : str, optional
        Name of dimensions to concat results along.
        Also Used by :class:`~lnpy.ensembles.xGrandCanonical`.
    concat_coords : string, optional
        parameters `coords `to :func:`xarray.concat`
    unstack : bool, default=True
        If True, then outputs will be unstacked using :meth:`xarray.DataArray.unstack`
    single_state : bool, default=True
        If True, verify that all data has same shape, and value of `state_kws`.
        That is, all ``lnpi`` are for a single state.
    *args **kwargs
        Extra arguments to Series constructor

    """

    _concat_dim = "sample"
    _concat_coords = "different"
    _use_joblib = True
    _xarray_output = True
    _xarray_unstack = True
    _xarray_dot_kws: Final = {"optimize": "optimal"}
    _use_cache = True

    def __init__(
        self,
        data: Sequence[lnPiMasked] | pd.Series[Any],
        index: ArrayLike | pd.Index[Any] | pd.MultiIndex | None = None,
        xarray_output: bool = True,
        concat_dim: str | None = None,
        concat_coords: str | None = None,
        unstack: bool = True,
        name: Hashable | None = None,
        base_class: str | type = "first",
        dtype: DTypeLike | None = None,
    ) -> None:
        if concat_dim is not None:
            self._concat_dim = concat_dim
        if concat_coords is not None:
            self._concat_coords = concat_coords
        if xarray_output is not None:
            self._xarray_output = xarray_output
        if unstack is not None:
            self._xarray_unstack = unstack
        # update index name:
        # self._series.index.name = self._concat_dim

        if isinstance(data, self.__class__):
            x = data
            data = x.s

        self._base_class = base_class
        self._verify = self._base_class is not None

        series: pd.Series[Any] = pd.Series(  # type: ignore[misc]
            data=data,
            index=index,  # type: ignore[arg-type]
            dtype=dtype,  # type: ignore[arg-type]
            name=name,
        )
        self._verify_series(series)
        self._series = series
        self._cache: dict[str, Any] = {}

    def _verify_series(self, series: pd.Series[Any]) -> None:
        if not self._verify:
            return

        base_class = self._base_class
        if isinstance(base_class, str) and base_class.lower() == "first":
            base_class = type(series.iloc[0])
        if isinstance(base_class, str):
            raise TypeError

        for d in series:
            if not issubclass(type(d), base_class):
                msg = f"all elements must be of type {base_class}"
                raise TypeError(msg)

        # lnpy
        first = series.iloc[0]
        state_kws = first.state_kws
        shape = first.shape
        # _base  = first._base

        for lnpi in series:
            if lnpi.state_kws != state_kws or lnpi.shape != shape:
                raise ValueError
            # would like to do this, but
            # fails for parallel builds
            # assert lnpi._base is _base

    def new_like(
        self,
        data: Sequence[lnPiMasked] | pd.Series[Any] | None = None,
        index: ArrayLike | pd.Index[Any] | pd.MultiIndex | None = None,
        **kwargs: Any,
    ) -> Self:
        """Create new object with optional new data/index"""
        if data is None:
            data = self.s

        return type(self)(
            data=data,
            index=index,
            xarray_output=self._xarray_output,
            concat_dim=self._concat_dim,
            concat_coords=self._concat_coords,
            unstack=self._xarray_unstack,
            **kwargs,
        )

    # ** Series Specific
    @property
    def series(self) -> pd.Series[Any]:
        """View of the underlying :class:`pandas.Series`"""
        return self._series

    @series.setter
    def series(self, series: pd.Series[Any]) -> None:
        self._cache = {}
        self._verify_series(series)
        self._series = series

    @property
    def s(self) -> pd.Series[Any]:
        """Alias to :meth:`series`"""
        return self.series

    def __iter__(self) -> Iterator[lnPiMasked]:
        return iter(self._series)  # pyright: ignore[reportCallIssue,reportArgumentType]

    @property
    def values(self) -> MyNDArray:
        """Series values"""
        return self._series.to_numpy()

    @property
    def items(self) -> MyNDArray:
        """Alias to :attr:`values`"""
        return self.values

    @property
    def index(self) -> pd.Index[Any]:
        """Series index"""
        return self._series.index

    @property
    def name(self) -> Hashable:
        """Series name"""
        return self._series.name

    def copy(self) -> Self:
        return type(self)(data=self.s, base_class=self._base_class)

    def _wrapped_pandas_method(
        self, mtd: str, wrap: bool = False, *args: Any, **kwargs: Any
    ) -> lnPiMasked | pd.Series[Any] | Self:
        """Wrap a generic pandas method to ensure it returns a GeoSeries"""
        val = getattr(self._series, mtd)(*args, **kwargs)
        if wrap and type(val) == pd.Series:
            val = self.new_like(val)
        return val  # type: ignore[no-any-return]

    def xs(
        self,
        key: Hashable | Sequence[Hashable],
        axis: int = 0,
        level: Hashable | Sequence[Hashable] | None = None,
        drop_level: bool = False,
        wrap: bool = True,
    ) -> Self | pd.Series[Any] | lnPiMasked:
        """Interface to :meth:`pandas.Series.xs`"""
        return self._wrapped_pandas_method(
            "xs", wrap=wrap, key=key, axis=axis, level=level, drop_level=drop_level
        )

    def __getitem__(self, key: Any) -> Self | lnPiMasked:
        """Interface to :meth:`pandas.Series.__getitem__`"""
        return self._wrapped_pandas_method("__getitem__", wrap=True, key=key)  # type: ignore[return-value]

    def __setitem__(
        self, idx: Any, values: lnPiMasked | Sequence[lnPiMasked] | pd.Series[Any]
    ) -> None:
        """Interface to :meth:`pandas.Series.__setitem__`"""
        self._series[idx] = values

    def __len__(self) -> int:
        return len(self.s)

    def append(
        self,
        to_append: pd.Series[Any] | Self,
        ignore_index: bool = False,
        verify_integrity: bool = True,
        concat_kws: Mapping[str, Any] | None = None,
        inplace: bool = False,
    ) -> Self:
        """
        Interface to :func:`pandas.concat`


        Parameters
        ----------
        to_append : object
            Object to append
        ignore_index : bool, default=False
        verify_integrity : bool, default=True
        concat_kws : mapping, optional
            Extra arguments to
        inplace : bool, default=False

        See Also
        --------
        pandas.concat
        """
        if hasattr(to_append, "series") and isinstance(to_append.series, pd.Series):
            series = to_append.series
        elif isinstance(to_append, pd.Series):
            series = to_append
        else:
            msg = f"Unknown to append type={type(to_append)}"
            raise ValueError(msg)

        if concat_kws is None:
            concat_kws = {}

        s = pd.concat(
            (self.series, series),
            ignore_index=ignore_index,
            verify_integrity=verify_integrity,
            **concat_kws,
        )

        if inplace:
            self.series = s
            return self

        return self.new_like(s)

    def droplevel(self, level: int | Hashable | Sequence[int | Hashable]) -> Self:
        """
        New object with dropped level

        See Also
        --------
        pandas.Series.droplevel
        """
        return self.new_like(self._series.droplevel(level=level, axis=0))  # type: ignore[arg-type,unused-ignore]

    def apply(
        self,
        func: Callable[..., Any],
        convert_dtype: bool = True,
        args: tuple[Any, ...] = (),
        wrap: bool = False,
        **kwds: Any,
    ) -> Self | pd.Series[Any]:
        """Interface to :meth:`pandas.Series.apply`"""

        return self._wrapped_pandas_method(  # type: ignore[return-value]
            "apply",
            wrap=wrap,
            func=func,
            convert_dtype=convert_dtype,
            args=args,
            **kwds,
        )

    def sort_index(self, *args: Any, **kwargs: Any) -> Self:
        """Interface to :meth:`pandas.Series.sort_index`"""
        return self._wrapped_pandas_method("sort_index", *args, wrap=True, **kwargs)  # type: ignore[misc,return-value]

    @overload
    def groupby(
        self,
        by: Hashable | Sequence[Hashable] = ...,
        *,
        level: int | Hashable | Sequence[int | Hashable] | None = ...,
        as_index: bool = ...,
        sort: bool = ...,
        group_keys: bool = ...,
        observed: bool = ...,
        dropna: bool = ...,
        wrap: Literal[False] = ...,
    ) -> SeriesGroupBy[Any, Any]: ...

    @overload
    def groupby(
        self,
        by: Hashable | Sequence[Hashable] = ...,
        *,
        level: int | Hashable | Sequence[int | Hashable] | None = ...,
        as_index: bool = ...,
        sort: bool = ...,
        group_keys: bool = ...,
        observed: bool = ...,
        dropna: bool = ...,
        wrap: Literal[True],
    ) -> _Groupby: ...

    @overload
    def groupby(
        self,
        by: Hashable | Sequence[Hashable] = ...,
        *,
        level: int | Hashable | Sequence[int | Hashable] | None = ...,
        as_index: bool = ...,
        sort: bool = ...,
        group_keys: bool = ...,
        observed: bool = ...,
        dropna: bool = ...,
        wrap: bool,
    ) -> SeriesGroupBy[Any, Any] | _Groupby: ...

    def groupby(
        self,
        by: Hashable | Sequence[Hashable] | None = None,
        *,
        level: int | Hashable | Sequence[int | Hashable] | None = None,
        as_index: bool = True,
        sort: bool = True,
        group_keys: bool = True,
        # squeeze=False,
        observed: bool = False,
        dropna: bool = True,
        wrap: bool = False,
    ) -> SeriesGroupBy[Any, Any] | _Groupby:
        """
        Wrapper around :meth:`pandas.Series.groupby`.

        Parameters
        ----------
        wrap : bool, default=False
            if True, try to wrap output in class of self

        See Also
        --------
        pandas.Series.groupby
        """
        group = self.s.groupby(  # type: ignore[call-overload]
            by=by,  # pyright: ignore[reportArgumentType]
            axis=0,
            level=level,
            as_index=as_index,
            sort=sort,
            group_keys=group_keys,
            # squeeze=squeeze,
            observed=observed,
            dropna=dropna,
        )

        if wrap:
            return _Groupby(self, group)
        return group  # type: ignore[no-any-return]

    @overload
    def groupby_allbut(
        self,
        drop: Hashable | list[Hashable] | tuple[Hashable],
        *,
        wrap: Literal[False] = ...,
        **kwargs: Any,
    ) -> SeriesGroupBy[Any, Any]: ...

    @overload
    def groupby_allbut(
        self,
        drop: Hashable | list[Hashable] | tuple[Hashable],
        *,
        wrap: Literal[True],
        **kwargs: Any,
    ) -> _Groupby: ...

    @overload
    def groupby_allbut(
        self,
        drop: Hashable | list[Hashable] | tuple[Hashable],
        *,
        wrap: bool,
        **kwargs: Any,
    ) -> SeriesGroupBy[Any, Any] | _Groupby: ...

    def groupby_allbut(
        self,
        drop: Hashable | list[Hashable] | tuple[Hashable],
        *,
        wrap: bool = False,
        **kwargs: Any,
    ) -> SeriesGroupBy[Any, Any] | _Groupby:
        """Groupby all but columns in drop"""
        from .utils import allbut

        if isinstance(drop, list):
            drop = tuple(drop)
        elif not isinstance(drop, tuple):
            drop = (drop,)

        by = allbut(self.index.names, *drop)  # pyright: ignore[reportArgumentType]

        # To suppress annoying errors.
        if len(by) == 1:
            return self.groupby(by=by[0], wrap=wrap, **kwargs)
        return self.groupby(by=by, wrap=wrap, **kwargs)

    @classmethod
    def _concat_to_series(
        cls,
        objs: Sequence[Self]
        | Sequence[pd.Series[Any]]
        | Mapping[Hashable, Self]
        | Mapping[Hashable, pd.Series[Any]],
        **concat_kws: Any,
    ) -> pd.Series[Any]:
        from collections.abc import Mapping, Sequence

        if isinstance(objs, Sequence):
            first = objs[0]
            if isinstance(first, cls):
                objs = tuple(x._series for x in objs)
        elif isinstance(objs, Mapping):
            out = {}
            remap = None
            for k in objs:
                v = objs[k]
                if remap is None:
                    remap = bool(isinstance(v, cls))
                if remap:
                    out[k] = v._series
                else:
                    out[k] = v
            objs = out
        else:
            msg = f"bad input type {type(objs[0])}"
            raise TypeError(msg)
        return pd.concat(objs, **concat_kws)  # type: ignore[return-value,arg-type]

    def concat_like(
        self,
        objs: Sequence[Self]
        | Sequence[pd.Series[Any]]
        | Mapping[Hashable, Self]
        | Mapping[Hashable, pd.Series[Any]],
        **concat_kws: Any,
    ) -> Self:
        """Concat a sequence of objects like `self`"""
        s = self._concat_to_series(objs, **concat_kws)
        return self.new_like(s)

    @classmethod
    def concat(
        cls,
        objs: Sequence[Self]
        | Sequence[pd.Series[Any]]
        | Mapping[Hashable, Self]
        | Mapping[Hashable, pd.Series[Any]],
        concat_kws: Mapping[str, Any] | None = None,
        *args: Any,
        **kwargs: Any,
    ) -> Self:
        """Create collection from sequence of objects"""
        if concat_kws is None:
            concat_kws = {}
        s = cls._concat_to_series(objs, **concat_kws)
        return cls(s, *args, **kwargs)

    # Note: use property(cached.meth(func)) here
    # normal cache.prop has some logic issues
    # with this pattern.
    @property
    @cached.meth
    def loc(self) -> _LocIndexer:
        return _LocIndexer(self)

    @property
    @cached.meth
    def iloc(self) -> _iLocIndexer:
        return _iLocIndexer(self)

    @property
    @cached.meth
    def query(self) -> _Query:
        return _Query(self)

    @property
    @cached.meth
    def zloc(self) -> _LocIndexer_unstack_zloc:
        return _LocIndexer_unstack_zloc(self)

    @property
    @cached.meth
    def mloc(self) -> _LocIndexer_unstack_mloc:
        return _LocIndexer_unstack_mloc(self)

    # ** lnPi Specific
    @cached.prop
    def _lnz_series(self) -> pd.Series[Any]:
        return self._series.apply(lambda x: x.lnz)  # type: ignore[no-any-return]

    def __repr__(self) -> str:
        return f"<class {self.__class__.__name__}>\n{self._lnz_series!r}"

    def __str__(self) -> str:
        return str(self._lnz_series)

    @property
    def state_kws(self) -> dict[str, Any]:
        """state_kws from first :class:`~lnpy.lnpidata.lnPiMasked`"""

        return self.iloc[0].state_kws

    @property
    def nlnz(self) -> int:
        """Number of unique lnzs"""
        return len(self.index.droplevel("phase").drop_duplicates())

    @cached.prop
    def index_frame(self) -> pd.DataFrame:
        """
        Values (from :class:`xarray.DataArray`) for each sample.

        includes a column 'lnz_index' which is the unique lnz values
        regardless of phase
        """
        sample_frame = (
            self.index.droplevel("phase")
            .drop_duplicates()
            .to_frame()
            .assign(lnz_sample=lambda x: np.arange(len(x)))["lnz_sample"]
        )
        return (
            self.index.to_frame()
            .reset_index("phase", drop=True)[["phase"]]
            .assign(lnz_index=lambda x: sample_frame[x.index])
            .reset_index()
        )

    @overload
    def _get_lnz(
        self, component: int, *, iloc: int = ..., zloc: int | None = ...
    ) -> float: ...

    @overload
    def _get_lnz(
        self, component: None = ..., *, iloc: int = ..., zloc: int | None = ...
    ) -> MyNDArray: ...

    def _get_lnz(
        self, component: int | None = None, *, iloc: int = 0, zloc: int | None = None
    ) -> float | MyNDArray:
        """
        Helper function to.
        returns self.iloc[idx].lnz[component]
        """
        v: lnPiMasked = (  # pyright: ignore[reportAssignmentType]
            (self.zloc[zloc]._series if zloc is not None else self._series).iloc[iloc]
        )
        lnz = v.lnz
        if component is not None:
            lnz = lnz[component]
        return lnz

    def _get_level(self, level: str = "phase") -> pd.Index[Any]:
        """Return level values from index"""
        index = self.index
        if isinstance(index, pd.MultiIndex):
            level_idx = index.names.index(level)
            index = index.levels[level_idx]
        return index

    def get_index_level(self, level: str = "phase") -> pd.Index[Any]:
        """Get index values for specified level"""
        return self._get_level(level=level)

    # @cached.prop
    @property
    def _nrec(self) -> int:
        return len(self._series)

    def _lnpi_tot(self, fill_value: float | None = None) -> MyNDArray:
        # old method
        # return np.stack([x.filled() for x in self])

        # new method
        # this is no faster than the original
        # but makes clear where the time is being spent
        first = self.iloc[0]
        n = len(self)
        out = np.empty((n, *first.shape), dtype=first.dtype)
        seq = get_tqdm((x.filled(fill_value) for x in self), total=n)
        for i, x in enumerate(seq):
            out[i, ...] = x
        return out

    # @property
    # def _lnpi_0_tot(self):
    #     return np.array([x.data.ravel()[0] for x in self])

    @property
    def _lnz_tot(self) -> MyNDArray:
        return np.stack([x.lnz for x in self])

    @property
    def lnz(self) -> MyNDArray:
        return self._lnz_tot

    def _pi_params(
        self, fill_value: float | None = None
    ) -> tuple[MyNDArray, MyNDArray, MyNDArray]:
        first = self.iloc[0]
        n = len(self)

        pi_norm = np.empty((n, *first.shape), dtype=first.dtype)
        pi_sum = np.empty(n, dtype=first.dtype)
        lnpi_zero = np.empty(n, dtype=first.dtype)

        seq = get_tqdm(
            (x._pi_params(fill_value) for x in self), total=n, desc="pi_norm"
        )

        for i, (_pi_norm, _pi_sum, _lnpi_zero) in enumerate(seq):
            pi_norm[i, ...] = _pi_norm
            pi_sum[i, ...] = _pi_sum
            lnpi_zero[i, ...] = _lnpi_zero
        return pi_norm, pi_sum, lnpi_zero

    def wrap_list_results(
        self, items: Sequence[Any] | Sequence[xr.DataArray]
    ) -> Sequence[Any] | xr.DataArray:
        """Utility to wrap output in :class:xarray.DataArray"""
        if self._xarray_output and isinstance(items[0], xr.DataArray):
            return xr.concat(items, self.index, coords=self._concat_coords)  # type: ignore[call-overload,no-any-return]
        return items

    ##################################################
    # Constructors
    @classmethod
    def from_list(
        cls,
        items: Sequence[lnPiMasked],
        index: Iterable[int] | MyNDArray,
        **kwargs: Any,
    ) -> Self:
        """
        Create collection from list of :class:`~lnpy.lnpidata.lnPiMasked` objects.

        Parameters
        ----------
        items : sequence of lnPiMasked
            Sequence of lnPi
        index : sequence
            Sequence of phases ID for each lnPi
        *args
            Extra positional arguments to `cls`
        **kwargs :
            Extra keyword arguments to `cls`

        Returns
        -------
        lnPiCollection
        """

        table = pd.DataFrame(
            [lnpi._index_dict(phase) for lnpi, phase in zip(items, index)]
        )
        new_index = pd.MultiIndex.from_frame(table)
        return cls(data=items, index=new_index, **kwargs)

    @classmethod
    def from_builder(
        cls,
        lnzs: Sequence[float] | MyNDArray,
        # TODO(wpk): make better type for build_phases.
        build_phases: Callable[..., tuple[list[lnPiMasked], MyNDArray]],
        ref: lnPiMasked | None = None,
        build_kws: Mapping[str, Any] | None = None,
        nmax: int | None = None,
        # concat_kws: Mapping[str, Any] |None=None,
        base_class: str | type = "first",
        **kwargs: Any,
    ) -> Self:
        """
        Build collection from scalar builder

        Parameters
        ----------
        lnzs : sequence of float
            One dimensional array of lnz value for the varying component.
        ref : lnPiMasked
            lnpi_phases to reweight to get list of lnpi's
        build_phases : callable
            Typically one of `PhaseCreator.build_phases_mu` or `PhaseCreator.build_phases_dmu`
        build_kws : optional
            optional arguments to `build_phases`

        Returns
        -------
        lnPiCollection


        See Also
        --------
        ~lnpy.segment.PhaseCreator.build_phases
        ~lnpy.segment.PhaseCreator.build_phases_mu
        ~lnpy.segment.PhaseCreator.build_phases_dmu

        """
        if build_kws is None:
            build_kws = {}

        build_kws = dict(build_kws, phases_factory=False)
        seq = get_tqdm(lnzs, desc="build")

        items: list[lnPiMasked] = []
        index: list[int] = []
        for data, idx in parallel_map(
            build_phases, seq, ref=ref, nmax=nmax, **build_kws
        ):
            items += data
            index += list(idx)
        return cls.from_list(items, index, base_class=base_class, **kwargs)

    ################################################################################
    # dataarray io
    def to_dataarray(
        self, dtype: DTypeLike | None = None, reset_index: bool = True
    ) -> xr.DataArray:
        """
        Convert collection to a :class:`~xarray.DataArray`

        Parameters
        ----------
        dtype : `numpy.dtype`, optional
            Default to `numpy.uint8`.
        reset_index : bool, default=True
        """

        dtype = dtype or np.uint8

        labels = []
        indexes = []

        for _, g in self.groupby_allbut("phase"):
            indexes.append(g.index[[0]])

            features = np.array(g.index.get_level_values("phase")) + 1
            masks = [x.mask for x in g]
            labels.append(
                masks_to_labels(masks, features=features, convention=False, dtype=dtype)
            )

        index = indexes[0].append(indexes[1:])  # type: ignore[no-untyped-call]

        data = np.stack(labels)

        out = (
            xr.DataArray(
                data,
                dims=self.xge.dims_rec + self.xge.dims_n,
                name="labels",
            )
            .assign_coords(**{self._concat_dim: index, **self.state_kws})
            .assign_attrs(**self.xge._standard_attrs)
        )

        if reset_index:
            out = out.reset_index(self._concat_dim)

        return out

    @classmethod
    @docfiller.decorate
    def from_labels(
        cls,
        ref: lnPiMasked,
        labels: Sequence[MyNDArray],
        lnzs: Sequence[float | MyNDArray] | MyNDArray,
        features: Sequence[int] | MyNDArray | None = None,
        include_boundary: bool = False,
        labels_kws: Mapping[str, Any] | None = None,
        check_features: bool = True,
        **kwargs: Any,
    ) -> Self:
        r"""
        Create from reference :class:`~lnpy.lnpidata.lnPiMasked` and labels array


        Parameters
        ----------
        ref : lnPiMasked
        labels : sequence of ndarray of int
            Each ``labels[i]`` is a labels array for each value of ``lnzs[i]``.
            That is, the labels for different phases at a given value of `lnz`.
        lnzs : sequence
            Each lnzs[i] will be passed to ``ref.reweight``.
        {features}
        {include_boundary}
        labels_kws : mapping, optional
        {check_features}
        **kwargs
            Extra arguments past to :meth:`from_list`

        See Also
        --------
        ~lnpy.utils.labels_to_masks
        from_list
        """

        if labels_kws is None:
            labels_kws = {}
        if len(labels) != len(lnzs):
            raise ValueError

        items = []
        indexes = []

        for label, lnz in zip(labels, lnzs):
            lnpi = ref.reweight(lnz)

            masks, features_tmp = labels_to_masks(
                labels=label,
                features=features,
                include_boundary=include_boundary,
                convention=False,
                check_features=check_features,
                **labels_kws,
            )

            index = list(np.array(features_tmp) - 1)
            items += lnpi.list_from_masks(masks, convention=False)
            indexes += index

        return cls.from_list(items=items, index=indexes, **kwargs)

    @classmethod
    @docfiller.decorate
    def from_dataarray(
        cls,
        ref: lnPiMasked,
        da: xr.DataArray,
        grouper: Hashable = "sample",
        include_boundary: bool = False,
        labels_kws: Mapping[str, Any] | None = None,
        features: Sequence[int] | MyNDArray | None = None,
        check_features: bool = True,
        **kwargs: Any,
    ) -> Self:
        """
        Create a collection from DataArray of labels

        Parameters
        ----------
        ref : lnPiMasked
        da : DataArray or int
            Labels.
        grouper : Hashable
            Name of dimension(s) to group along to give a single label array
        {features}
        {check_features}


        See Also
        --------
        from_labels
        """

        labels = []
        lnzs = []

        for _, g in da.groupby(grouper):
            lnzs.append(np.array([g.coords[k] for k in da.attrs["dims_lnz"]]))
            labels.append(g.values)

        return cls.from_labels(
            ref=ref,
            labels=labels,
            lnzs=lnzs,
            labels_kws=labels_kws,
            features=features,
            include_boundary=include_boundary,
            check_features=check_features,
            **kwargs,
        )  # yapf: disable

    @cached.prop
    @docfiller.decorate
    def xge(self) -> ensembles.xGrandCanonical:
        """{accessor.xge}"""
        from .ensembles import xGrandCanonical

        return xGrandCanonical(self)

    @cached.prop
    def wfe(self) -> lnpienergy.wFreeEnergyCollection:
        """Accessor to :class:`~lnpy.lnpienergy.wFreeEnergyPhases` from :attr:`wfe_phases`."""
        from .lnpienergy import wFreeEnergyCollection

        return wFreeEnergyCollection(self)

    @cached.prop
    def wfe_phases(self) -> lnpienergy.wFreeEnergyPhases:
        """Accessor to :class:`~lnpy.lnpienergy.wFreeEnergyCollection` from :attr:`wfe`."""
        from .lnpienergy import wFreeEnergyPhases

        return wFreeEnergyPhases(self)

    @property
    def wlnPi(self) -> lnpienergy.wFreeEnergyCollection:  # noqa: N802
        """
        Deprecated accessor to :class:`~lnpy.lnpienergy.wFreeEnergyCollection` from :attr:`wlnPi`.

        Alias to :attr:`wfe`
        """
        warn(
            "Using `wlnPi` accessor is deprecated.  Please use `wfe` accessor instead",
            stacklevel=1,
        )
        return self.wfe

    @property
    def wlnPi_single(self) -> lnpienergy.wFreeEnergyPhases:  # noqa: N802
        """
        Deprecated accessor to :class:`~lnpy.lnpienergy.wFreeEnergyPhases` from :attr:`wlnPi_single`.

        Alias to :attr:`wfe_phases`
        """
        warn(
            "Using `wlnPi_single is deprecated.  Please use `self.wfe_phases` instead",
            stacklevel=1,
        )
        return self.wfe_phases

    @cached.prop
    def spinodal(self) -> stability.Spinodals:
        """Accessor to :class:`~lnpy.stability.Spinodals`"""
        from .stability import Spinodals

        return Spinodals(self)

    @cached.prop
    def binodal(self) -> stability.Binodals:
        """Accessor to :class:`~lnpy.stability.Binodals`"""
        from .stability import Binodals

        return Binodals(self)

    def stability_append(
        self,
        other: Self | None,
        append: bool = True,
        sort: bool = True,
        copy_stability: bool = True,
    ) -> Self:
        if (not append) and (not copy_stability):
            msg = "one of append or copy_stability must be True"
            raise ValueError(msg)

        if other is None:
            other = self
        spin = other.spinodal
        bino = other.binodal
        if append:
            new = self.append(spin.appender).append(bino.appender)  # type: ignore[arg-type]
            if sort:
                new = new.sort_index()
        else:
            new = self.copy()
        if copy_stability:
            # TODO(wpk): fix this hack
            new._cache["spinodal"] = spin
            new._cache["binodal"] = bino
            # new.spinodal = spin
            # new.binodal = bino
        return new


################################################################################
# Accessors for ColleectionlnPi
# lnPiCollection.register_accessor("xge", xge_accessor)
# lnPiCollection.register_accessor("wfe", wfe_accessor)
# lnPiCollection.register_accessor("wfe_phases", wfe_phases_accessor)
# lnPiCollection.register_accessor("wlnPi", wlnPi_accessor)
# lnPiCollection.register_accessor("wlnPi_single", wlnPi_single_accessor)



import numpy as np
import pytest

from lnpy.lnpicollectionutils import build_grid


def test_build_grid0() -> None:
    x0 = -0.2124

    grid = build_grid(
        x=None,
        dx=0.5,
        x0=x0,
        x_range=(-2, 2),
        even_grid=True,
        digits=1,
        outlier=False,
    )

    assert grid == pytest.approx(np.arange(-2, +2 + 0.25, 0.5))


def test_build_grid1() -> None:
    x0 = -0.2124

    grid = build_grid(
        x=None,
        dx=0.5,
        x0=x0,
        offsets=(-2, +2),
        even_grid=True,
        digits=1,
        outlier=False,
    )

    assert grid == pytest.approx(np.arange(-2, +2, 0.5))



"""
Segmentation of lnPi (:mod:`~lnpy.segment`)
===========================================
Routines to segment lnPi

 1. find max/peaks in lnPi
 2. segment lnPi about these peaks
 3. determine free energy difference between segments
    a. Merge based on low free energy difference
 4. combination of 1-3.
"""

from __future__ import annotations

import warnings
from collections.abc import Iterable
from functools import lru_cache
from typing import TYPE_CHECKING, cast, overload  # , TypedDict

import numpy as np
from module_utilities.docfiller import DocFiller

from .docstrings import docfiller
from .lnpienergy import wFreeEnergy
from .lnpiseries import lnPiCollection

if TYPE_CHECKING:
    from typing import Any, Literal, Mapping, Sequence

    from numpy.typing import ArrayLike

    from ._typing import (
        MyNDArray,
        PeakError,
        PeakStyle,
        PhasesFactorySignature,
        TagPhasesSignature,
    )
    from .lnpidata import lnPiMasked


# * Common doc strings
_docstrings_local = r"""
Parameters
----------
data : array-like
    Image data to analyze
min_distance : int or sequence of int, optional
    min_distance parameter.  If sequence, then call
    :func:`~skimage.feature.peak_local_max` until number of peaks ``<=num_peaks_max``.
    Default value is ``(5, 10, 15, 20, 25)``.
connectivity_morphology | connectivity : int, optional
    Maximum number of orthogonal hops to consider a pixel/voxel as a neighbor.
    Accepted values are ranging from 1 to ``input.ndim``. If ``None``, a full
    connectivity of ``input.ndim`` is used.
connectivity_watershed | connectivity : ndarray, optional
    An array with the same number of dimensions as image whose non-zero
    elements indicate neighbors for connection. Following the scipy convention,
    default is a one-connected array of the dimension of the image.
num_peaks_max : int, optional
    Max number of maxima/peaks to find. If not specified, any number of peaks allowed.
peak_style | style : {'indices', 'mask', 'marker'}
    Controls output style

    * indices : indices of peaks
    * mask : array of bool marking peak locations
    * marker : array of int
markers : int, or ndarray of int, optional
    Same shape as image. The desired number of markers, or an array marking the
    basins with the values to be assigned in the label matrix. Zero means not a
    marker. If None (no markers given), the local minima of the image are used
    as markers.

lnz_buildphases_mu | lnz : list of float or None
    list with one element equal to None.  This is the component which will be varied
    For example, lnz=[lnz0, None, lnz2] implies use values of lnz0,lnz2 for components 0 and 2, and
    vary component 1.
dlnz_buildphases_dmu | dlnz : list of float or None
    list with one element equal to None.  This is the component which will be varied
    For example, dlnz=[dlnz0,None,dlnz2] implies use values of dlnz0,dlnz2
    for components 0 and 2, and vary component 1.
    dlnz_i = lnz_i - lnz_index, where lnz_index is the value varied.
phase_creator : :class:`PhaseCreator`
    Factory method to create phases collection object.
    For example, :meth:`.lnPiCollection.from_list`.
phases_factory : callable or bool, default=True
    Function to convert list of phases into Phases object.
    If `phases_factory` ``True``, revert to `self.phases_factory`.
    If `phases_factory` is ``False``, do not apply a factory, and
    return list of :class:`lnpy.lnpidata.lnPiMasked` and array of phase indices.
"""


docfiller_local = docfiller.append(
    DocFiller.from_docstring(_docstrings_local, combine_keys="parameters")
).decorate


@overload
def peak_local_max_adaptive(
    data: MyNDArray,
    *,
    mask: MyNDArray | None = ...,
    min_distance: Sequence[int] | None = ...,
    style: Literal["indices"] = ...,
    threshold_rel: float = ...,
    threshold_abs: float = ...,
    num_peaks_max: int | None = ...,
    connectivity: int | None = ...,
    errors: PeakError = ...,
    **kwargs: Any,
) -> tuple[MyNDArray, ...]: ...


@overload
def peak_local_max_adaptive(
    data: MyNDArray,
    *,
    mask: MyNDArray | None = ...,
    min_distance: Sequence[int] | None = ...,
    style: Literal["mask", "marker"],
    threshold_rel: float = ...,
    threshold_abs: float = ...,
    num_peaks_max: int | None = ...,
    connectivity: int | None = ...,
    errors: PeakError = ...,
    **kwargs: Any,
) -> MyNDArray: ...


@overload
def peak_local_max_adaptive(
    data: MyNDArray,
    *,
    mask: MyNDArray | None = ...,
    min_distance: Sequence[int] | None = ...,
    style: str,
    threshold_rel: float = ...,
    threshold_abs: float = ...,
    num_peaks_max: int | None = ...,
    connectivity: int | None = ...,
    errors: PeakError = ...,
    **kwargs: Any,
) -> MyNDArray | tuple[MyNDArray, ...]: ...


@docfiller_local
def peak_local_max_adaptive(
    data: MyNDArray,
    *,
    mask: MyNDArray | None = None,
    min_distance: Sequence[int] | None = None,
    style: PeakStyle | str = "indices",
    threshold_rel: float = 0.0,
    threshold_abs: float = 0.2,
    num_peaks_max: float | None = None,
    connectivity: int | None = None,
    errors: PeakError = "warn",
    **kwargs: Any,
) -> MyNDArray | tuple[MyNDArray, ...]:
    """
    Find local max with fall backs min_distance and filter.

    This is an adaptation of :func:`~skimage.feature.peak_local_max`, which is
    called iteratively until the number of `peaks` is less than `num_peaks_max`.

    Parameters
    ----------
    {data}
    {mask_image}
    {min_distance}
    {peak_style}
    threshold_rel, threshold_abs : float
        thresholds parameters
    {num_peaks_max}
    {connectivity_morphology}
    errors : {{'ignore','raise','warn'}}, default='warn'
        - If raise, raise exception if npeaks > num_peaks_max
        - If ignore, return all found maxima
        - If warn, raise warning if npeaks > num_peaks_max
    **kwargs
        Extra arguments to :func:`~skimage.feature.peak_local_max`

    Returns
    -------
    out : array of int or list of array of bool
        Depending on the value of `indices`.

    Notes
    -----
    The option `mask` is passed as the value `labels` in :func:`~skimage.feature.peak_local_max`

    See Also
    --------
    ~skimage.feature.peak_local_max
    ~skimage.morphology.label
    """
    import bottleneck
    from skimage.feature import peak_local_max
    from skimage.morphology import label as morphology_label

    possible_styles = {"indices", "mask", "marker"}
    if style not in possible_styles:
        msg = f"{style=} not in {possible_styles}"
        raise ValueError(msg)

    if min_distance is None:
        min_distance = [5, 10, 15, 20, 25]

    if num_peaks_max is None:
        num_peaks_max = np.inf

    if not isinstance(min_distance, Iterable):
        min_distance = [min_distance]

    data = data - bottleneck.nanmin(data)
    kwargs = dict({"exclude_border": False}, **kwargs)

    n = idx = None
    for md in min_distance:
        idx = peak_local_max(
            data,
            min_distance=md,
            labels=mask,
            threshold_abs=threshold_abs,
            threshold_rel=threshold_rel,
            # this option removed in future
            **kwargs,
        )

        n = len(idx)
        if n <= num_peaks_max:
            break

    if n is None or idx is None:
        msg = "failed to find peaks"
        raise ValueError(msg)

    if n > num_peaks_max:
        if errors == "ignore":
            pass
        elif errors in {"raise", "ignore"}:
            message = f"{n} maxima found greater than {num_peaks_max}"
            if errors == "raise":
                raise RuntimeError(message)
            warnings.warn(message, stacklevel=1)

    idx = tuple(idx.T)
    if style == "indices":
        return cast("tuple[MyNDArray, ...]", idx)

    out = np.zeros_like(data, dtype=bool)
    out[idx] = True

    if style == "marker":
        out = morphology_label(out, connectivity=connectivity)
    return cast("MyNDArray", out)


# TODO(wpk): create these for below
# class Peak_Kws(TypedDict, total=False):
#     min_distance: Sequence[int]
#     style: PeakStyle
#     threshold_rel: float
#     threshold_abs: float
#     num_peaks_max: int
#     connectivity: int
#     errors: PeakError
#     # peak_local_max options:
#     exclude_border: int | tuple[int, ...] | bool
#     num_peaks: int
#     footprint: MyNDArray
#     p_norm: float

# class Watershed_Kws(TypedDict, total=False):


@docfiller_local
class Segmenter:
    """
    Data segmenter:


    Parameters
    ----------
    {min_distance}
    {peak_kws}
    {watershed_kws}
    """

    def __init__(
        self,
        peak_kws: Mapping[str, Any] | None = None,
        watershed_kws: Mapping[str, Any] | None = None,
    ) -> None:
        if peak_kws is None:
            peak_kws = {}
        self.peak_kws = peak_kws

        if watershed_kws is None:
            watershed_kws = {}
        self.watershed_kws = watershed_kws

    @overload
    def peaks(
        self,
        data: MyNDArray,
        mask: MyNDArray | None = ...,
        *,
        num_peaks_max: int | None = ...,
        style: Literal["marker"] = ...,
        **kwargs: Any,
    ) -> MyNDArray: ...

    @overload
    def peaks(
        self,
        data: MyNDArray,
        mask: MyNDArray | None = ...,
        *,
        num_peaks_max: int | None = ...,
        style: Literal["mask"],
        **kwargs: Any,
    ) -> MyNDArray: ...

    @overload
    def peaks(
        self,
        data: MyNDArray,
        mask: MyNDArray | None = ...,
        *,
        num_peaks_max: int | None = ...,
        style: Literal["indices"],
        **kwargs: Any,
    ) -> tuple[MyNDArray, ...]: ...

    @overload
    def peaks(
        self,
        data: MyNDArray,
        mask: MyNDArray | None = ...,
        *,
        num_peaks_max: int | None = ...,
        style: str,
        **kwargs: Any,
    ) -> MyNDArray | tuple[MyNDArray, ...]: ...

    @docfiller_local
    def peaks(
        self,
        data: MyNDArray,
        mask: MyNDArray | None = None,
        *,
        num_peaks_max: int | None = None,
        style: PeakStyle | str = "marker",
        **kwargs: Any,
    ) -> MyNDArray | tuple[MyNDArray, ...]:
        """
        Interface to :func:`peak_local_max_adaptive` with default values from `self`.


        Parameters
        ----------
        {data}
        {num_peaks_max}
        {peak_style}
        {mask_image}
        **kwargs
            Extra arguments to :func:`peak_local_max_adaptive`.

        Returns
        -------
        ndarray of int or sequence of ndarray
            If ``style=='marker'``, then return label array.  Otherwise,
            return indices of peaks.

        Notes
        -----
        Any value not set will be inherited from `self.peak_kws`


        See Also
        --------
        peak_local_max_adaptive

        """

        if mask is not None:
            kwargs["mask"] = mask
        if num_peaks_max is not None:
            kwargs["num_peaks_max"] = num_peaks_max
        kwargs["style"] = style
        kwargs = dict(self.peak_kws, **kwargs)
        return peak_local_max_adaptive(data, **kwargs)  # type: ignore[no-any-return]

    @docfiller_local
    def watershed(
        self,
        data: MyNDArray,
        markers: int | MyNDArray,
        mask: MyNDArray,
        connectivity: int | MyNDArray | None = None,
        **kwargs: Any,
    ) -> MyNDArray:
        """
        Interface to :func:`skimage.segmentation.watershed` function

        Parameters
        ----------
        {data}
        {markers}
        {mask_image}
        {connectivity_watershed}
        **kwargs
            Extra arguments to :func:`~skimage.segmentation.watershed`

        Returns
        -------
        {labels}

        See Also
        --------
        ~skimage.segmentation.watershed
        """
        from skimage.segmentation import watershed

        if connectivity is None:
            connectivity = data.ndim

        kwargs = dict(self.watershed_kws, connectivity=connectivity, **kwargs)
        return watershed(data, markers=markers, mask=mask, **kwargs)  # type: ignore[no-any-return]

    @docfiller_local
    def segment_lnpi(
        self,
        lnpi: lnPiMasked,
        markers: int | MyNDArray | None = None,
        find_peaks: bool = True,
        num_peaks_max: int | None = None,
        connectivity: MyNDArray | None = None,
        peaks_kws: Mapping[str, Any] | None = None,
        watershed_kws: Mapping[str, Any] | None = None,
    ) -> MyNDArray:
        """
        Perform segmentations of lnPi object using watershed on negative of lnPi data.

        Parameters
        ----------
        lnpi : lnPiMasked
            Object to be segmented
        {markers}

        find_peaks : bool, default=True
            If True, use :func:`peak_local_max_adaptive` to construct `markers`.
        {num_peaks_max}
        {connectivity_watershed}
        {peak_kws}
        {watershed_kws}

        Returns
        -------
        {labels}


        See Also
        --------
        Segmenter.watershed
        ~skimage.segmentation.watershed

        """

        if markers is None:
            if find_peaks:
                if peaks_kws is None:
                    peaks_kws = {}
                else:
                    peaks_kws = dict(peaks_kws)
                    peaks_kws.pop("style", None)

                markers = self.peaks(
                    lnpi.data,
                    mask=~lnpi.mask,
                    num_peaks_max=num_peaks_max,
                    connectivity=connectivity,
                    style="marker",
                    **peaks_kws,
                )
            else:
                markers = num_peaks_max

        if not isinstance(markers, (int, np.ndarray)):
            msg = f"{type(markers)=} must be int or np.ndarray"
            raise TypeError(msg)

        if watershed_kws is None:
            watershed_kws = {}
        return self.watershed(
            -lnpi.data, markers=markers, mask=~lnpi.mask, connectivity=connectivity
        )


class PhaseCreator:
    """
    Helper class to create phases

    Parameters
    ----------
    nmax : int
        Maximum number of phases to allow
    nmax_peak : int, optional
        if specified, the allowable number of peaks to locate.
        This can be useful for some cases.  These phases will be merged out at the end.
    ref : lnPiMasked, optional
        Reference object.
    segmenter : :class:`Segmenter`, optional
        segmenter object to create labels/masks. Defaults to using base segmenter.
    segment_kws : mapping, optional
        Optional arguments to be passed to :meth:`.Segmenter.segment_lnpi`.
    tag_phases : callable, optional
        Optional function which takes a list of :class:`~.lnPiMasked` objects
        and returns on integer label for each object.
    phases_factory : callable, optional
        Factory function for returning Collection from a list of :class:`~lnpy.lnpidata.lnPiMasked` object.
        Defaults to :meth:`.lnPiCollection.from_list`.
    free_energy_kws : mapping, optional
        Optional arguments to ...
    merge_kws : mapping, optional
        Optional arguments to :func:`.merge_regions`

    """

    def __init__(
        self,
        nmax: int,
        nmax_peak: int | None = None,
        ref: lnPiMasked | None = None,
        segmenter: Segmenter | None = None,
        segment_kws: Mapping[str, Any] | None = None,
        tag_phases: TagPhasesSignature | None = None,
        phases_factory: PhasesFactorySignature | None = None,
        free_energy_kws: Mapping[str, Any] | None = None,
        merge_kws: Mapping[str, Any] | None = None,
    ) -> None:
        self.nmax = nmax
        self.ref = ref

        self.segmenter = segmenter or Segmenter()
        self.tag_phases = tag_phases
        self.phases_factory = phases_factory or lnPiCollection.from_list

        self.segment_kws = {} if segment_kws is None else dict(segment_kws)
        self.segment_kws["num_peaks_max"] = nmax_peak or nmax * 2

        self.free_energy_kws = free_energy_kws or {}
        self.merge_kws = (
            {}
            if merge_kws is None
            else dict(merge_kws, convention=False, nfeature_max=self.nmax)
        )

    # TODO(wpk): make this work with integer or string phase_ids
    @staticmethod
    def _merge_phase_ids(
        ref: lnPiMasked,
        phase_ids: Sequence[int] | MyNDArray,
        lnpis: list[lnPiMasked],
    ) -> tuple[MyNDArray, list[lnPiMasked]]:
        """Perform merge of phase_ids/index"""
        from scipy.spatial.distance import pdist

        phase_ids = np.asarray(phase_ids)
        if len(phase_ids) == 1:
            # only single phase_id
            return phase_ids, lnpis

        dist = pdist(phase_ids.reshape(-1, 1)).astype(int)
        if not np.any(dist == 0):
            # all different
            return phase_ids, lnpis

        phase_ids_new = []
        masks_new = []
        for idx in np.unique(phase_ids):
            where = np.where(idx == phase_ids)[0]
            mask = np.all([lnpis[i].mask for i in where], axis=0)

            phase_ids_new.append(idx)
            masks_new.append(mask)
        lnpis_new = ref.list_from_masks(masks_new, convention=False)

        return np.asarray(phase_ids_new), lnpis_new

    @overload
    def build_phases(
        self,
        lnz: float | Sequence[float] | ArrayLike | MyNDArray | None = ...,
        ref: lnPiMasked | None = ...,
        *,
        efac: float | None = ...,
        nmax: int | None = ...,
        nmax_peak: int | None = ...,
        connectivity: int | None = ...,
        reweight_kws: Mapping[str, Any] | None = ...,
        merge_phase_ids: bool = True,
        merge_phases: bool = True,
        phases_factory: PhasesFactorySignature | Literal[True] = ...,
        phase_kws: Mapping[str, Any] | None = ...,
        segment_kws: Mapping[str, Any] | None = ...,
        free_energy_kws: Mapping[str, Any] | None = ...,
        merge_kws: Mapping[str, Any] | None = ...,
        tag_phases: TagPhasesSignature | None = ...,
    ) -> lnPiCollection: ...

    @overload
    def build_phases(
        self,
        lnz: float | Sequence[float] | ArrayLike | MyNDArray | None = ...,
        ref: lnPiMasked | None = ...,
        *,
        efac: float | None = ...,
        nmax: int | None = ...,
        nmax_peak: int | None = ...,
        connectivity: int | None = ...,
        reweight_kws: Mapping[str, Any] | None = ...,
        merge_phase_ids: bool = True,
        merge_phases: bool = True,
        phases_factory: Literal[False],
        phase_kws: Mapping[str, Any] | None = ...,
        segment_kws: Mapping[str, Any] | None = ...,
        free_energy_kws: Mapping[str, Any] | None = ...,
        merge_kws: Mapping[str, Any] | None = ...,
        tag_phases: TagPhasesSignature | None = ...,
    ) -> tuple[list[lnPiMasked], MyNDArray]: ...

    @overload
    def build_phases(
        self,
        lnz: float | Sequence[float] | ArrayLike | MyNDArray | None = ...,
        ref: lnPiMasked | None = ...,
        *,
        efac: float | None = ...,
        nmax: int | None = ...,
        nmax_peak: int | None = ...,
        connectivity: int | None = ...,
        reweight_kws: Mapping[str, Any] | None = ...,
        merge_phase_ids: bool = True,
        merge_phases: bool = True,
        phases_factory: PhasesFactorySignature | bool,
        phase_kws: Mapping[str, Any] | None = ...,
        segment_kws: Mapping[str, Any] | None = ...,
        free_energy_kws: Mapping[str, Any] | None = ...,
        merge_kws: Mapping[str, Any] | None = ...,
        tag_phases: TagPhasesSignature | None = ...,
    ) -> tuple[list[lnPiMasked], MyNDArray] | lnPiCollection: ...

    @docfiller_local
    def build_phases(
        self,
        lnz: float | Sequence[float] | ArrayLike | MyNDArray | None = None,
        ref: lnPiMasked | None = None,
        *,
        efac: float | None = None,
        nmax: int | None = None,
        nmax_peak: int | None = None,
        connectivity: int | None = None,
        reweight_kws: Mapping[str, Any] | None = None,
        merge_phase_ids: bool = True,
        merge_phases: bool = True,
        phases_factory: PhasesFactorySignature | bool = True,
        phase_kws: Mapping[str, Any] | None = None,
        segment_kws: Mapping[str, Any] | None = None,
        free_energy_kws: Mapping[str, Any] | None = None,
        merge_kws: Mapping[str, Any] | None = None,
        tag_phases: TagPhasesSignature | None = None,
    ) -> tuple[list[lnPiMasked], MyNDArray] | lnPiCollection:
        """
        Construct 'phases' for a lnPi object.

        This is quite an involved process.  The steps are

        * Optionally find the location of the maxima in lnPi.
        * Segment lnpi using watershed
        * Merge phases which are energetically similar
        * Optionally merge phases which have the same `phase_id`

        Parameters
        ----------
        lnz : int or sequence of int, optional
            lnz value to evaluate `ref` at.  If not specified, use
            `ref.lnz`
        ref : lnPiMasked
            Object to be segmented
        efac : float, optional
            Optional value to use in energetic merging of phases.
        nmax : int, optional
            Maximum number of phases.  Defaults to `self.nmax`
        nmax_peak : int, optional
            Maximum number of peaks to allow in :func:`peak_local_max_adaptive`.
            Note that this value can be larger than `nmax`.  Defaults to `self.nmax_peak`.
        {connectivity_morphology}
        reweight_kws : mapping, optional
            Extra arguments to `ref.reweight`
        merge_phase_ids : bool, default=True
            If True and calling `tag_phases` routine, merge phases with same phase_id.
        {phases_factory}
        phase_kws : mapping, optional
            Extra arguments to `phases_factory`
        segment_kws : mapping, optional
            Extra arguments to `self.segmenter.segment_lnpi`
        free_energy_kws : mapping, optional
            Extra arguments to free energy calculation
        merge_kws : mapping, optional
            Extra arguments to merge
        tag_phases : callable, optional
            Function to tag phases.  Defaults to `self.tag_phases`

        Returns
        -------
        output : list of lnPiMasked and ndarray, or lnPiCollection
            If no phase creator, return list of lnPiMasked objects and array of phase indices.
            Otherwise, lnPiCollection object.
        """

        def _combine_kws(
            class_kws: Mapping[str, Any] | None,
            passed_kws: Mapping[str, Any] | None,
            **default_kws: Any,
        ) -> dict[str, Any]:
            return dict(class_kws or {}, **default_kws, **(passed_kws or {}))

        if ref is None:
            if self.ref is None:
                msg = "must specify ref or self.ref"
                raise ValueError(msg)
            ref = self.ref

        # reweight
        if lnz is not None:
            if reweight_kws is None:
                reweight_kws = {}
            ref = ref.reweight(lnz, **reweight_kws)

        if nmax is None:
            nmax = self.nmax

        if nmax_peak is None:
            nmax_peak = nmax * 2

        connectivity_kws = {}
        if connectivity is not None:
            connectivity_kws["connectivity"] = connectivity

        if nmax > 1:
            # segment lnpi using watershed
            segment_kws = _combine_kws(
                self.segment_kws,
                segment_kws,
                num_peaks_max=nmax_peak,
                **connectivity_kws,
            )
            labels = self.segmenter.segment_lnpi(lnpi=ref, **segment_kws)

            # analyze w = - lnPi
            free_energy_kws = _combine_kws(
                self.free_energy_kws, free_energy_kws, **connectivity_kws
            )
            wlnpi = wFreeEnergy.from_labels(
                data=ref.data, labels=labels, **free_energy_kws
            )

            if merge_phases:
                # merge
                other_kws = {} if efac is None else {"efac": efac}
                merge_kws = _combine_kws(
                    self.merge_kws, merge_kws, nfeature_max=nmax, **other_kws
                )
                masks, _, _ = wlnpi.merge_regions(**merge_kws)
            else:
                masks = wlnpi.masks

            # list of lnpi
            lnpis = ref.list_from_masks(masks, convention=False)

            # tag phases?
            if tag_phases is None:
                tag_phases = self.tag_phases
            if tag_phases is not None:
                index = tag_phases(lnpis)
                if merge_phase_ids:
                    index, lnpis = self._merge_phase_ids(ref, index, lnpis)
            else:
                index = list(range(len(lnpis)))
        else:
            lnpis = [ref]
            index = [0]

        if isinstance(phases_factory, bool):
            if phases_factory:
                phases_factory = self.phases_factory
            else:
                return lnpis, np.asarray(index)

        phase_kws = phase_kws or {}
        return phases_factory(items=lnpis, index=index, **phase_kws)

    def build_phases_mu(self, lnz: list[float | None]) -> BuildPhases_mu:
        """
        Factory constructor at fixed values of `mu`

        Parameters
        ----------
        {lnz_buildphases_mu}

        See Also
        --------
        BuildPhases_mu

        Examples
        --------
        >>> import lnpy.examples
        >>> e = lnpy.examples.hsmix_example()

        The default build phases from this multicomponent system requires specifies the
        activity for both species.  For example:

        >>> e.phase_creator.build_phases([0.1, 0.2])
        <class lnPiCollection>
        lnz_0  lnz_1  phase
        0.1    0.2    0        [0.1, 0.2]
                      1        [0.1, 0.2]
        dtype: object


        But if we want to creat phases at a fixed value of either lnz_0 or lnz_1, we can
        do the following:

        >>> b = e.phase_creator.build_phases_mu([None, 0.5])

        Note the syntax [None, 0.5].  This means that calling `b(lnz_0)` will
        create a new object at [lnz_0, 0.5].

        >>> b(0.1)
        <class lnPiCollection>
        lnz_0  lnz_1  phase
        0.1    0.5    0        [0.1, 0.5]
                      1        [0.1, 0.5]
        dtype: object

        Likewise, we can fix lnz_0 with

        >>> b = e.phase_creator.build_phases_mu([0.5, None])

        >>> b(0.1)
        <class lnPiCollection>
        lnz_0  lnz_1  phase
        0.5    0.1    0        [0.5, 0.1]
                      1        [0.5, 0.1]
        dtype: object


        To create an object at fixed value of ``dmu_i = lnz_i - lnz_fixed``, we use the following:

        >>> b = e.phase_creator.build_phases_dmu([None, 0.5])

        Now any phase created will have ``lnz = [lnz_0, 0.5 + lnz_0]``

        >>> b(0.5)
        <class lnPiCollection>
        lnz_0  lnz_1  phase
        0.5    1.0    0        [0.5, 1.0]
                      1        [0.5, 1.0]
        dtype: object

        """
        return BuildPhases_mu(lnz, self)

    def build_phases_dmu(self, dlnz: list[float | None]) -> BuildPhases_dmu:
        """
        Factory constructor at fixed values of `dmu`.

        Parameters
        ----------
        {dlnz_buildphases_dmu}


        See Also
        --------
        BuildPhases_dmu
        build_phases_mu
        """
        return BuildPhases_dmu(dlnz, self)


class BuildPhasesBase:
    """Base class to build Phases objects from scalar values of `lnz`."""

    def __init__(self, x: list[float | None], phase_creator: PhaseCreator) -> None:
        self._phase_creator = phase_creator
        self._set_x(x)

    @property
    def x(self) -> list[float | None]:
        return self._x

    @x.setter
    def x(self, x: list[float | None]) -> None:
        self._set_x(x)

    @property
    def phase_creator(self) -> PhaseCreator:
        return self._phase_creator

    def _set_x(self, x: list[float | None]) -> None:
        # assert sum([x is None for x in x]) == 1
        if sum([x is None for x in x]) != 1:
            msg = f"{x=} must have a single element which is None.  This will be the dimension varied."
            raise ValueError(msg)
        self._x = x
        self._ncomp = len(self._x)
        self._index = self._x.index(None)
        self._set_params()

    @property
    def index(self) -> int:
        """Index number which varies"""
        return self._index

    def _set_params(self) -> None:
        pass

    def _get_lnz(self, lnz_index: float) -> MyNDArray:
        # to be implemented in child class
        raise NotImplementedError

    @overload
    def __call__(
        self,
        lnz_index: float,
        *,
        phases_factory: PhasesFactorySignature | Literal[True] = ...,
        **kwargs: Any,
    ) -> lnPiCollection: ...

    @overload
    def __call__(
        self,
        lnz_index: float,
        *,
        phases_factory: Literal[False],
        **kwargs: Any,
    ) -> tuple[list[lnPiMasked], MyNDArray]: ...

    @overload
    def __call__(
        self,
        lnz_index: float,
        *,
        phases_factory: PhasesFactorySignature | bool,
        **kwargs: Any,
    ) -> tuple[list[lnPiMasked], MyNDArray] | lnPiCollection: ...

    def __call__(
        self,
        lnz_index: float,
        *,
        phases_factory: PhasesFactorySignature | bool = True,
        **kwargs: Any,
    ) -> tuple[list[lnPiMasked], MyNDArray] | lnPiCollection:
        """
        Build phases from scalar value of lnz.

        Parameters
        ----------
        lnz_index : float
            Value of lnz for `self.index` index.
        {phases_factory}
        **kwargs
            Extra arguments to :meth:`PhaseCreator.build_phases`

        Returns
        -------
        output : list of lnPiMasked and ndarray, or lnPiCollection
            If no phase creator, return list of lnPiMasked objects and array of phase indices.
            Otherwise, lnPiCollection object.

        See Also
        --------
        PhaseCreator.build_phases
        """
        lnz = self._get_lnz(lnz_index)
        return self._phase_creator.build_phases(
            lnz=lnz, phases_factory=phases_factory, **kwargs
        )


# from .utils import get_lnz_iter
@docfiller_local
class BuildPhases_mu(BuildPhasesBase):  # noqa: N801
    """
    create phases from scalar value of mu for fixed value of mu for other species

    Parameters
    ----------
    {lnz_buildphases_mu}
    {phase_creator}
    """

    def __init__(self, lnz: list[float | None], phase_creator: PhaseCreator) -> None:
        super().__init__(x=lnz, phase_creator=phase_creator)

    def _get_lnz(self, lnz_index: float) -> MyNDArray:
        lnz = self.x.copy()
        lnz[self.index] = lnz_index
        return np.asarray(lnz)


@docfiller_local
class BuildPhases_dmu(BuildPhasesBase):  # noqa: N801
    """
    Create phases from scalar value of mu at fixed value of dmu for other species

    Parameters
    ----------
    {dlnz_buildphases_dmu}
    {phase_creator}
    """

    def __init__(self, dlnz: list[float | None], phase_creator: PhaseCreator) -> None:
        super().__init__(x=dlnz, phase_creator=phase_creator)

    def _set_params(self) -> None:
        self._dlnz: MyNDArray = np.array([x if x is not None else 0.0 for x in self.x])

    def _get_lnz(self, lnz_index: float) -> MyNDArray:
        return self._dlnz + lnz_index


@lru_cache(maxsize=10)
def get_default_phasecreator(nmax: int) -> PhaseCreator:
    return PhaseCreator(nmax=nmax)
